package com.rapidminer.mapreduce;

import java.util.Collection;

/**
 * <p>MapReduce is a framework to compute on large data in parallel. 
 * The data is divided in chunks and will be processed in parallel by a
 * map-function independently from all over chunks. Every map-function
 * yields an interim result which will be collected and reduced by a 
 * reduce-function into a single result.</p>
 * 
 * <p>The map-function is represented by the {@link Mapper} class and the 
 * reduce-function by the {@link Reducer} respectively.</p>
 * 
 * @author Marcin Skirzynski
 *
 * @param <I>	the input class for the map-function
 * @param <M>	the interim result class
 * @param <O>	the output class for the result
 */
public interface MapReduceExecutor<I, M, O> {

	/**
	 * <p>Distributes the given chunks to instances of the {@link Mapper} class
	 * which are created by the given factory. The interim results will be collected
	 * and passed to the {@link Reducer} created by the other factory which reduces
	 * all the result into a single one which will be returned by this function.</p>
	 *  
	 * @param chunks			a collection of chunks which will be distributed
	 * @param mapperFactory		the factory for the {@link Mapper} class
	 * @param reducerFactory	the factory for the {@link Reducer} class
	 * @return					the result from the reducer
	 * @throws MapReduceException for any exception caught within the MapReduce context
	 */
	public O mapReduce(	Collection<I> chunks, 
						MapperFactory<I, M> mapperFactory,
						ReducerFactory<M, O> reducerFactory)
					throws MapReduceException;
	
}
