package com.rapidminer.mapreduce;

import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * Implementation of a {@link MapReduceExecutor} which process all of the map- and
 * reduce steps locally on a single machine using multiple threads for parallel
 * processing.
 * 
 * @author Marcin Skirzynski
 *
 * @param <I>	the type of the input
 * @param <M>	the type of the interim result
 * @param <O>	the type of the output
 */
public class MapReduceExecutorLocal<I, M, O> implements MapReduceExecutor<I, M, O> {
	
	/**
	 * Number of threads which will be used for parallelization
	 */
	private final int nrOfThreads;

	/**
	 * Creates a local executor with the given number of threads
	 *  
	 * @param nrOfThreads	number of threads which will be used for parallelization
	 */
	public MapReduceExecutorLocal(int nrOfThreads) {
		this.nrOfThreads = nrOfThreads;
	}
	
	/**
	 * Creates an local executor which determines the available CPUs and creates
	 * an appropriate amount of threads. 
	 */
	public MapReduceExecutorLocal() {
		this(Runtime.getRuntime().availableProcessors());
	}

	/**
	 * Creates a thread for every data chunk, executes them in parallel
	 * and reduces the interim results.
	 * {@inheritDoc}
	 */
	@Override
	public O mapReduce(	Collection<I> chunks, 
						MapperFactory<I, M> mapperFactory, 
						ReducerFactory<M, O> reducerFactory) 
					throws MapReduceException {
		
		try {
			
			// Executor for the mapper threads and ...
			ExecutorService executor = Executors.newFixedThreadPool(nrOfThreads);
			// ... the completion service for receiving the results.
			CompletionService<M> interimResultPool = new  ExecutorCompletionService<M>(executor);

			/*
			 * Create mapper for every chunk and add them as callable
			 * to the completion service to compute in a thread.
			 */
			int maps = 0;
			for( I input : chunks ) {
				Mapper<I, M> mapper = mapperFactory.createMapper();
				interimResultPool.submit(new MapperCallable(mapper, input));
				maps++;
			}
			// Doesn't accept any other threads and shuts down if the last thread is read
			executor.shutdown();
			
			// Collect the interim results ...
			Collection<M> intermediateResults = new ArrayList<M>();
			while(maps-->0) {
				intermediateResults.add(interimResultPool.take().get());
			}

			// .. and use a reducer to reduce these results.
			Reducer<M, O> reducer = reducerFactory.createReducer();
			return reducer.reduce(intermediateResults);
			
		} catch (Exception e) {
			if(MapReduceException.class.isInstance(e))
				throw (MapReduceException)e;
			else
				throw new MapReduceException(e);
		}
	}
	
	/**
	 * {@link Callable} class for executing a mapper in a thread.
	 *  
	 * @author Marcin Skirzynski
	 *
	 */
	private class MapperCallable implements Callable<M> {
		
		/**
		 * The {@link Mapper} which will be executed in a thread
		 */
		private final Mapper<I, M> mapper;
		
		/**
		 * The input for the {@link Mapper}
		 */
		private final I input;

		/**
		 * Constructor for a {@link Callable} which invoices the {@link Mapper}.
		 * 
		 * @param mapper	the {@link Mapper} to be executed in a thread
		 * @param input		the input for the {@link Mapper}
		 */
		public MapperCallable(Mapper<I, M> mapper, I input) {
			this.mapper = mapper;
			this.input = input;
		}

		/**
		 * Executes the {@link Mapper} and returns the result as
		 * a result of the thread.
		 */
		@Override
		public M call() throws Exception {
			return mapper.map(input);
		}
		
	}
	
	

}
