package com.rapidminer.operator.meta;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.set.SplittedExampleSet;
import com.rapidminer.mapreduce.MapReduceException;
import com.rapidminer.mapreduce.MapReduceExecutor;
import com.rapidminer.mapreduce.MapReduceExecutorLocal;
import com.rapidminer.mapreduce.Mapper;
import com.rapidminer.mapreduce.MapperFactory;
import com.rapidminer.mapreduce.Reducer;
import com.rapidminer.mapreduce.ReducerFactory;
import com.rapidminer.operator.ExecutionUnit;
import com.rapidminer.operator.IOObject;
import com.rapidminer.operator.IOObjectCollection;
import com.rapidminer.operator.OperatorChain;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.MDTransformationRule;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.conditions.BooleanParameterCondition;

/**
 * <p>Operator which mimics the MapReduce-approach.</p>
 * 
 * <p>The operator chain has two nested execution units. The first one
 * represents the map-function ({@link ExecutionUnitMapper}) and the second 
 * the reduce-function ({@link ExecutionUnitReducer}). The {@link ExampleSet} 
 * input will be partitioned into  </p>
 *  
 * @author Marcin Skirzynski
 *
 */
public class MapReduceOperator 
						extends OperatorChain 
						implements 
							MapperFactory<ExampleSet, IOObject>,
							ReducerFactory<IOObject, IOObject> {
	
	/**
	 * Parameter key for the number of threads which shall be used.
	 */
	private static final String PK_NR_OF_THREADS = "nr_of_threads";
	
	/**
	 * Parameter key for the number of chunks which will be created.
	 */
	private static final String PK_NR_OF_CHUNKS = "nr_of_chunks";
	
	/**
	 * Parameter key for the boolean if the number of chunks shall be the number of threads.
	 */
	private static final String PK_CHUNKS_EQUAL_THREADS = "chunks_equal_threads";
	
	//input
	protected final InputPort input = getInputPorts().createPort("exampleSet", ExampleSet.class);
	protected final InputPort additional = getInputPorts().createPort("additional");
	
	// output
	protected final OutputPort output = getOutputPorts().createPort("output");
	
	// map
	protected final OutputPort mapInput = getMapExecutionUnit().getInnerSources().createPort("exampleSet");
	protected final OutputPort additionalMapInput = getMapExecutionUnit().getInnerSources().createPort("additional");
	protected final InputPort mapOutput = getMapExecutionUnit().getInnerSinks().createPort("interim result");
	
	// reduce
	protected final OutputPort reduceInput = getReducerExecutionUnit().getInnerSources().createPort("interim result");
	protected final OutputPort additionalReducerInput = getReducerExecutionUnit().getInnerSources().createPort("additional");
	protected final InputPort reduceOutput = getReducerExecutionUnit().getInnerSinks().createPort("output");
	
	/**
	 * Additional {@link IOObject} which will be passed to the map- and reduce-step.
	 */
	private IOObject additionalIOObject;
	
	/**
	 * The id of the last execution unit created and 0 if non unit was
	 * created yet.
	 */
	private static AtomicInteger lastExecutionUnitId = new AtomicInteger(0);
	
	/**
	 * Constructor for the operator with generation of meta data rules
	 * 
	 * @param description
	 */
	public MapReduceOperator(OperatorDescription description) {
		super(description, "Map", "Reduce");
		
		getTransformer().addGenerationRule(reduceInput, IOObjectCollection.class);
		
		getTransformer().addRule(new MDTransformationRule() {
			@Override
			public void transformMD() {
				MetaData additionalMD = additional.getMetaData();
				if(additionalMD!=null) {
					additionalMapInput.deliverMD(additionalMD);
					additionalReducerInput.deliverMD(additionalMD);
				}
			}
		});
		getTransformer().addRule(new MDTransformationRule() {
			@Override
			public void transformMD() {
				MetaData outputMD = reduceOutput.getMetaData();
				if(outputMD!=null) {
					output.deliverMD(outputMD);
				}
			}
		});
	}

	@Override
	public void doWork() throws OperatorException {
		
		// Get parameters
		int nrOfThreads = getParameterAsInt(PK_NR_OF_THREADS);
		int nrOfChunks = getParameterAsBoolean(PK_CHUNKS_EQUAL_THREADS) ? nrOfThreads : getParameterAsInt(PK_NR_OF_CHUNKS);
		
		// Get inputs
		ExampleSet inputExampleSet = input.getData(ExampleSet.class);
		additionalIOObject = additional.getDataOrNull();
		
		// Collect chunks of data TODO do this in an other execution unit?!
		SplittedExampleSet splitted = 
				new SplittedExampleSet(inputExampleSet, nrOfChunks, SplittedExampleSet.LINEAR_SAMPLING, false, -1);
		Collection<ExampleSet> mapReduceInput = new ArrayList<ExampleSet>();
		for(int i=0; i<nrOfChunks; i++) {
			splitted.selectSingleSubset(i);
			ExampleSet exampleSet = (ExampleSet) splitted.copy();
			mapReduceInput.add(exampleSet);
		}
		
		// Start map and reduce!
		try {
			
			MapReduceExecutor<ExampleSet, IOObject, IOObject> executor = 
					new MapReduceExecutorLocal<ExampleSet, IOObject, IOObject>(nrOfThreads);
			IOObject result = executor.mapReduce(mapReduceInput, this, this);
			output.deliver(result);
			
		} catch (MapReduceException e) {
			throw new UserError(this, e, 935);
		} 
	}
	
	/**
	 * Returns the execution unit for the map-step.
	 * 
	 * @return	execution unit for the {@link Mapper}
	 */
	private ExecutionUnit getMapExecutionUnit() {
		return getSubprocess(0);
	}
	
	/**
	 * Returns the executions unit for the reduce-step
	 * 
	 * @return	execution unit for the {@link Reducer}
	 */
	private ExecutionUnit getReducerExecutionUnit() {
		return getSubprocess(1);
	}
	
	/**
	 * Returns the additional {@link IOObject} for the
	 * map- and reduce-step.
	 * 
	 * @return	additional {@link IOObject}
	 */
	private IOObject getAdditionalIOObject() {
		return additionalIOObject;
	}
	
	/**
	 * Adding parameters as the number of threads
	 */
	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		
		types.add(new ParameterTypeInt(PK_NR_OF_THREADS, 
				"Determines the number of threads which will be used for parallel execution",
				1, 											// Min
				Integer.MAX_VALUE, 							// Max
				Runtime.getRuntime().availableProcessors() 	// Default
		));
		
		types.add(new ParameterTypeBoolean(PK_CHUNKS_EQUAL_THREADS, "Indicates if the number of chunks wil be equal to the number of threads", true));
		
		ParameterTypeInt chunksType = new ParameterTypeInt(PK_NR_OF_CHUNKS,
				"The number of chunks to split the example set",
				1,
				Integer.MAX_VALUE,
				Runtime.getRuntime().availableProcessors()
			);
		
		chunksType.registerDependencyCondition(new BooleanParameterCondition(this, PK_CHUNKS_EQUAL_THREADS, true, false));
		types.add(chunksType);
		
		return types;
	}
	
	/**
	 * Inline class for which uses the first {@link ExecutionUnit} to create
	 * a {@link Mapper}.
	 * 
	 * @author Marcin Skirzynski
	 */
	public class ExecutionUnitMapper implements Mapper<ExampleSet, IOObject> {
		
		/**
		 * Executes the {@link ExecutionUnit} as the map-step.
		 */
		@Override
		public IOObject map(ExampleSet input) throws MapReduceException {
			try {
				int nextId = lastExecutionUnitId.addAndGet(1);
				System.err.println("Computing on "+input.size()+" examples with MapUnit#"+nextId);
				ExecutionUnit unit = cloneExecutionUnit(nextId);
				unit.getInnerSources().getPortByIndex(0).deliver(input);
				if(getAdditionalIOObject()!=null)
					unit.getInnerSources().getPortByIndex(1).deliver(getAdditionalIOObject());
				unit.execute();
				System.err.println("MapUnit#" + nextId + " finished");
				return unit.getInnerSinks().getPortByIndex(0).getData();
			} catch (OperatorException e) {
				throw new MapReduceException(e);
			}
		}

		/**
		 * Clones the {@link ExecutionUnit} for parallel execution.
		 * @return
		 */
		private ExecutionUnit cloneExecutionUnit(int id) {
			ExecutionUnit unit = new ExecutionUnit(getMapExecutionUnit().getEnclosingOperator(), "MapUnit#"+ id);
			
			for( OutputPort o : getMapExecutionUnit().getInnerSources().getAllPorts() ) {
				unit.getInnerSources().createPort(o.getName());
			}
			for( InputPort i : getMapExecutionUnit().getInnerSinks().getAllPorts() ) {
				unit.getInnerSinks().createPort(i.getName());
			}
			
			unit.cloneExecutionUnitFrom(getMapExecutionUnit(), true);
			return unit;
		}
		
	}
	
	/**
	 * Inline class for the {@link ExecutionUnit} which will be used for
	 * the reduce-step.
	 * 
	 * @author Marcin Skirzynski
	 *
	 */
	public class ExecutionUnitReducer implements Reducer<IOObject, IOObject> {

		/**
		 * Reducing the collection of {@link IOObject} to a single {@link IOObject}.
		 */
		@Override
		public IOObject reduce(Collection<IOObject> input) {
			IOObjectCollection<IOObject> iooCollection = new IOObjectCollection<IOObject>();
			for(IOObject ioObject : input ) {
				iooCollection.add(ioObject);
			}
			try {
				getReducerExecutionUnit().getInnerSources().getPortByIndex(0).deliver(iooCollection);
				if(getAdditionalIOObject()!=null)
					getReducerExecutionUnit().getInnerSources().getPortByIndex(1).deliver(getAdditionalIOObject());
				getReducerExecutionUnit().execute();
				return getReducerExecutionUnit().getInnerSinks().getPortByIndex(0).getData();
			} catch (OperatorException e) {
				e.printStackTrace();
				throw new RuntimeException(e);
			}
		}

	}

	/**
	 * Factory method for the {@link Reducer}.
	 */
	@Override
	public Reducer<IOObject, IOObject> createReducer() {
		return new ExecutionUnitReducer();
	}

	/**
	 * Factory method for the {@link Mapper}.
	 */
	@Override
	public Mapper<ExampleSet, IOObject> createMapper() {
		return new ExecutionUnitMapper();
	}
	

}
