package com.rapidminer.operator.performance;


import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import sun.management.ManagementFactory;

import com.rapidminer.Process;
import com.rapidminer.operator.ResultObjectAdapter;
import com.sun.management.OperatingSystemMXBean;

/**
 * Object which starts and stops the measuring of the currenct performance.
 * 
 * @author Marcin Skirzynski
 */
public class ComputingPerformance extends ResultObjectAdapter implements Runnable{
	private static final long serialVersionUID = 1342200442890323919L;
	
	private long lastSystemTime = 0;
	private long lastProcessCpuTime = 0;
	
	/**
	 * The process for this measurement
	 */
	private final Process process;
	
	/**
	 * The executor service for starting and stopping this as a thread.
	 */
	private ScheduledExecutorService service;

	/**
	 * The file where the values will be stored
	 */
	private final File file;

	/**
	 * The pooling interval for measuring the performance
	 */
	private final int interval;

	/**
	 * The writer for the file which has to be closed when stopping.
	 */
	private BufferedWriter out;
	
	/**
	 * Creates a performance object
	 */
	public ComputingPerformance(Process process, File file, int interval) {
		this.process = process;
		this.file = file;
		this.interval = interval;
	}
	
	/**
	 * Starts the measurement creating a new file and starting a thread for measurement.
	 */
	public void start() {
	
		try {
			if(file.exists()) {
				file.delete();
			}
			file.createNewFile();
			
			String header = "time;cpu;memory;operator\n";
			FileWriter fstream = new FileWriter(file,false);
			out = new BufferedWriter(fstream);
			out.write(header);
			out.flush();
		} catch (IOException e) {
			System.err.println("Can not open file - no performance will be measured");
			return;
		}
		
		/*
		 * Starting the thread
		 */
		service = Executors.newScheduledThreadPool(1);
		service.scheduleAtFixedRate(this, 1, interval, TimeUnit.SECONDS);
		
	}
	
	/**
	 * Stops the measurement by stopping the thread and closing the file 
	 */
	public void stop() {
		if(service!=null)
			service.shutdown();
		if(out!=null)
			try {
				out.close();
			} catch (IOException e) {
				System.err.println("Can not close file for performance measurement");
			}
	}
	
	/**
	 * The bean for determining the system infop 
	 */
	private OperatingSystemMXBean getSunOperatingSystemMXBean() {
		return (OperatingSystemMXBean) ManagementFactory.getOperatingSystemMXBean();
	}

	/**
	 * Returns the cpu usage since the last call of this method.
	 */
	public double getCpuUsage() {
		if (lastSystemTime == 0) {
			lastSystemTime = System.nanoTime();
			lastProcessCpuTime = getSunOperatingSystemMXBean().getProcessCpuTime();
			return 0.0;
		}

		long systemTime = System.nanoTime();
		long processCpuTime = 0;

		processCpuTime = getSunOperatingSystemMXBean().getProcessCpuTime();
		double cpuUsage = (double) (processCpuTime - lastProcessCpuTime)
				/ (systemTime - lastSystemTime);

		lastSystemTime = systemTime;
		lastProcessCpuTime = processCpuTime;

		return cpuUsage;
	}

	/**
	 * Returns the current used memory from the heap.
	 */
	public long getMemoryUsage() {
		return java.lang.management.ManagementFactory.getMemoryMXBean().getHeapMemoryUsage().getUsed()/1000000;
	}

	/**
	 * Single performance measurement.
	 */
	@Override
	public void run() {
		String operator = "unknown";
		if(process!=null) {
			operator = process.getCurrentOperator().getName();
		}
		
		String line = System.currentTimeMillis()+ ";" +(int)(getCpuUsage()*100) + ";" + getMemoryUsage() + ";" +operator+ "\n";
		try {
			out.write(line);
			out.flush();
		} catch (IOException e) {
			System.err.println("Can not write to file for performance measurement");
		}
	}
	
}
