package com.rapidminer.operator.performance;

import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import com.rapidminer.Process;
import com.rapidminer.ProcessListener;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.Value;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeInt;

/**
 * Provides the currently used memory as value for RapidMiner
 * which can be logged by the Log-Operator.
 * 
 * @author Marcin Skirzynski
 */
public class MemoryLog extends Operator implements Runnable {
	
	private static final String PK_POOLING_INTERVAL = "pooling-interval";

	/**
	 * The maximum amount of memory used.
	 */
	private long maxMemory = 0;
	
	/**
	 * Executor for threads which polls for memory consumption
	 */
	private ScheduledExecutorService service;
	
	/**
	 * Adds the memory as value
	 */
	public MemoryLog(OperatorDescription description) {
		super(description);
		
		addValue(new Value("Max-Memory", "The maximal memory used") {
			@Override
			public boolean isNominal() {
				return true;
			}
			@Override
			public Object getValue() {
				return Long.toString(maxMemory);
			}
		});
	}
	
	/**
	 * Starts thread to poll memory.
	 */
	@Override
	public void doWork() throws OperatorException {
		if(service==null) {
			service = Executors.newScheduledThreadPool(1);
			service.scheduleAtFixedRate(this, 1, getParameterAsInt(PK_POOLING_INTERVAL), TimeUnit.SECONDS);
			getProcess().getRootOperator().addProcessListener(new ProcessListener() {
				@Override
				public void processStarts(Process process) {}
				@Override
				public void processStartedOperator(Process process, Operator op) {}
				@Override
				public void processFinishedOperator(Process process, Operator op) {}
				@Override
				public void processEnded(Process process) {
					if(service!=null)
						service.shutdown();
					process.getRootOperator().removeProcessListener(this);
				}
			});
		}
	}

	/**
	 * Updates memory variable.
	 */
	@Override
	public void run() {
		long currentlyUsedMemory = getCurrentlyUsedMemory();
		if(maxMemory<currentlyUsedMemory)
			maxMemory = currentlyUsedMemory;
	}
	
	/**
	 * Returns the currently used memory.
	 */
	private long getCurrentlyUsedMemory() {
		return Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
	}
	
	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		
		types.add(new ParameterTypeInt(PK_POOLING_INTERVAL, "The interval in which the memory will be measured (in seconds)", 1, Integer.MAX_VALUE));
		
		return types;
	}

}
