package com.rapidminer.operator.preprocessing.association;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.learner.associations.BooleanAttributeItem;
import com.rapidminer.operator.learner.associations.FrequentItemSet;
import com.rapidminer.operator.learner.associations.FrequentItemSets;
import com.rapidminer.operator.learner.associations.Item;
import com.rapidminer.operator.learner.associations.fpgrowth.FPGrowth;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.ExampleSetPrecondition;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeDouble;
import com.rapidminer.tools.Ontology;

/**
 * Operator to calculate the {@link FrequentItemSets} for Single-Item sets. This operator
 * should be used instead {@link FPGrowth} with "max-item" set to 1, because {@link FPGrowth}
 * builds up an FP-Tree regardless of the itemset size. So for big data sets this isn't 
 * feasible. Goal should be to provide a patch for RapidMiner in {@link FPGrowth} to change
 * this behaviour.
 * 
 * There are a few code parts copied and pasted from the {@link FPGrowth} algorithm.
 * 
 * @author Sebastian Land, Marcin Skirzynski
 */
public class FrequentItem extends Operator {
	
	private static final String PARAMETER_MIN_SUPPORT = "min_support";

	private final InputPort exampleSetInput = getInputPorts().createPort("example set");

	private final OutputPort exampleSetOutput = getOutputPorts().createPort("example set");
	private final OutputPort frequentSetsOutput = getOutputPorts().createPort("frequent sets");

	public FrequentItem(OperatorDescription description) {
		super(description);
		exampleSetInput.addPrecondition(new ExampleSetPrecondition(exampleSetInput, Ontology.BINOMINAL));
		getTransformer().addGenerationRule(frequentSetsOutput, FrequentItemSets.class);
		getTransformer().addPassThroughRule(exampleSetInput, exampleSetOutput);
	}
	
	@Override
	public void doWork() throws OperatorException {
		
		ExampleSet exampleSet = exampleSetInput.getData(ExampleSet.class);
		int minTotalSupport = (int) Math.ceil(getParameterAsDouble(PARAMETER_MIN_SUPPORT) * exampleSet.size());

		// pre-computing data properties
		ExampleSet workingSet = preprocessExampleSet(exampleSet);

		// determining attributes and their positive indices
		Attribute[] attributes = new Attribute[workingSet.getAttributes().size()];
		double[] positiveIndices = new double[workingSet.getAttributes().size()];
		int i = 0;
		for (Attribute attribute : workingSet.getAttributes()) {
			attributes[i] = attribute;
			positiveIndices[i] = attribute.getMapping().getPositiveIndex();
			i++;
		}

		// map attributes to items
		Map<Attribute, Item> itemMapping = getAttributeMapping(workingSet);
		// computing frequency of 1-Item Sets
		getItemFrequency(workingSet, attributes, positiveIndices, itemMapping);

		FrequentItemSets sets = new FrequentItemSets(workingSet.size());
		for(Item item : itemMapping.values()) {
			if(item.getFrequency()>=minTotalSupport) {
				FrequentItemSet set = new FrequentItemSet();
				set.addItem(item, item.getFrequency());
				sets.addFrequentSet(set);
			}
		}
		
		exampleSetOutput.deliver(exampleSet);
		frequentSetsOutput.deliver(sets);
	}
	
	private ExampleSet preprocessExampleSet(ExampleSet exampleSet) {
		// precomputing data properties
		ExampleSet workingSet = (ExampleSet) exampleSet.clone();

		// remove unusuable attributes
		int oldAttributeCount = workingSet.getAttributes().size();
		removeNonBooleanAttributes(workingSet);
		int newAttributeCount = workingSet.getAttributes().size();
		if (oldAttributeCount != newAttributeCount) {
			int removeCount = oldAttributeCount - newAttributeCount;
			String message = null;
			if (removeCount == 1)
				message = "Removed 1 non-binominal attribute, frequent item set mining is only supported for the positive values of binominal attributes.";
			else
				message = "Removed " + removeCount + " non-binominal attributes, frequent item set mining is only supported for the positive values of binominal attributes.";
			logWarning(message);
		}

		return workingSet;
	}

	/**
	 * Removes every non boolean attribute.
	 * 
	 * @param exampleSet
	 *            exampleSet, which attributes are tested
	 */
	private void removeNonBooleanAttributes(ExampleSet exampleSet) {
		// removing non boolean attributes
		Collection<Attribute> deleteAttributes = new ArrayList<Attribute>();
		for (Attribute attribute : exampleSet.getAttributes()) {
			if (!attribute.isNominal() || (attribute.getMapping().size() != 2)) {
				deleteAttributes.add(attribute);
			}
		}
		for (Attribute attribute : deleteAttributes) {
			exampleSet.getAttributes().remove(attribute);
		}
	}

	/**
	 * This method maps the attributes of the given exampleSet to an Item.
	 * 
	 * @param exampleSet
	 *            the exampleSet which attributes are mapped
	 */
	private Map<Attribute, Item> getAttributeMapping(ExampleSet exampleSet) {
		// computing Attributes to test, because only boolean attributes are used
		Map<Attribute, Item> mapping = new HashMap<Attribute, Item>();
		for (Attribute attribute : exampleSet.getAttributes()) {
			mapping.put(attribute, new BooleanAttributeItem(attribute));
		}
		return mapping;
	}

	/**
	 * This method scans the exampleSet and counts the frequency of every item
	 * 
	 * @param exampleSet
	 *            the exampleSet to be scaned
	 * @param mapping
	 *            the mapping of attributes to items
	 */
	private void getItemFrequency(ExampleSet exampleSet, Attribute[] attributes, double[] positiveIndices, Map<Attribute, Item> mapping) {
		// iterate over exampleSet, counting item frequency
		for (Example currentExample : exampleSet) {
			int i = 0;
			for (Attribute attribute : attributes) {
				// if attribute is boolean and if attribute is the positive one --> increase frequency of item
				if (currentExample.getValue(attribute) == positiveIndices[i]) {
					mapping.get(attribute).increaseFrequency();
				}
				i++;
			}
		}
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.add(new ParameterTypeDouble(PARAMETER_MIN_SUPPORT, "The minimal support necessary in order to be a frequent item (set).", 0.0d, 1.0d, 0.95d));
		return types;
	}
}
