package com.rapidminer.operator.preprocessing.association;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.rapidminer.operator.IOObjectCollection;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.learner.associations.FrequentItemSet;
import com.rapidminer.operator.learner.associations.FrequentItemSets;
import com.rapidminer.operator.learner.associations.Item;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeDouble;

/**
 * Aggregates a collection of frequent item sets to a single set where all
 * frequencies for the same item sets are accumulated.
 * 
 * @author Marcin Skirzynski
 *
 */
public class FrequentItemSetsAggregation extends Operator {

	
	public static final String PARAMETER_MIN_SUPPORT = "min_support";

	/**
	 * The input is a collection of frequent item sets
	 */
	private InputPort itemSetsInputPort = getInputPorts().createPort("Collection of frequent itemsets", IOObjectCollection.class);
	
	/**
	 * Returns a single frequent item set
	 */
	private OutputPort itemSetOutputPort = getOutputPorts().createPort("Frequent Itemsets");
	
	/**
	 * Constructs this operator
	 * @param description
	 */
	public FrequentItemSetsAggregation(OperatorDescription description) {
		super(description);
		getTransformer().addGenerationRule(itemSetOutputPort, FrequentItemSets.class);
	}
	
	/**
	 * Do the merging of the sets.
	 */
	@Override
	public void doWork() throws OperatorException {
		IOObjectCollection<FrequentItemSets> itemSetCollection = itemSetsInputPort.getData();

		/*
		 *  Iterate over all found item sets and build a new item set
		 *  which doesn't contains duplicates.
		 */
		int nrOfTransactions = 0;
		Map<String, FrequentItemSet> uniqueItemSets = new HashMap<String, FrequentItemSet>();
		for(FrequentItemSets itemSets : itemSetCollection.getObjects()) {
			nrOfTransactions += itemSets.getNumberOfTransactions();
			for(FrequentItemSet itemSet : itemSets) {
				/*
				 * Checking if this item set was already found by generating a string
				 * which is unqiue for a set with the same items. The getItemsAsString
				 * does not work here because the items are sorted their according to their
				 * frequencies.
				 */
				String asString = getItemSetKey(itemSet);
				if(uniqueItemSets.containsKey(asString)) {
					// Create a new set with the accumulated frequencies
					uniqueItemSets.put(asString, merge(itemSet, uniqueItemSets.get(asString)));
				} else {
					uniqueItemSets.put(asString, itemSet);
				}
			}
		}
		
		FrequentItemSets result = new FrequentItemSets(nrOfTransactions);
		for(FrequentItemSet uniqueItemSet : uniqueItemSets.values()) {
			result.addFrequentSet(uniqueItemSet);
		}
		
		itemSetOutputPort.deliver(result);
		
	}

	/**
	 * Merges the given two frequent itemsets by creating a new set with new items
	 * and the accumulated frequencies. 
	 */
	private FrequentItemSet merge(FrequentItemSet fis1, FrequentItemSet fis2) {
		ArrayList<Item> items = new ArrayList<Item>();
		for(Item i1 : fis1.getItems()) {
			for(Item i2 : fis2.getItems()) {
				if(equalItemsExcludingFrequency(i1, i2)) {
					items.add(new StringItem(i1.toString(), i1.getFrequency() + i2.getFrequency()));
				}
			}
		}
		return new FrequentItemSet(items, fis1.getFrequency() + fis2.getFrequency());
	}
	
	/**
	 * Test if the two given Items are equal besides their frequency. Depends
	 * on the toString() method.
	 */
	private boolean equalItemsExcludingFrequency(Item i1, Item i2) {
		/*
		 * All current implementations of an Item returning just the name of
		 * this item. Since there is no other method, this is the only way without
		 * changing the Item-API of RapidMiner.
		 */
		return i1.toString().equals(i2.toString());
	}
	
	/**
	 * Gets a string for the items in this set so that every item set with the
	 * same items gets the same string without considering the frequency of the items
	 * as the getItemsAsString-method of the set does.
	 */
	private String getItemSetKey(FrequentItemSet itemSet) {
		List<String> itemStrings = new ArrayList<String>();
		for(Item item : itemSet.getItems()) {
			itemStrings.add(item.toString());
		}
		Collections.sort(itemStrings);
		StringBuilder builder = new StringBuilder();
		for(String s : itemStrings) {
			builder.append("$_");
			builder.append(s);
		}
		return builder.toString();
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.add(new ParameterTypeDouble(PARAMETER_MIN_SUPPORT, "The minimal support necessary in order to be added as a frequent item (set).", 0.0d, 1.0d, 0.0d));
		return types;
	}
	
}
