package com.rapidminer.operator.preprocessing.association;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Attributes;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.learner.associations.FrequentItemSet;
import com.rapidminer.operator.learner.associations.FrequentItemSets;
import com.rapidminer.operator.learner.associations.Item;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.ExampleSetPrecondition;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeDouble;
import com.rapidminer.tools.Ontology;

/**
 * From an aggregated frequent itemsets all false positives
 * will be pruned. The frequencies may be still to low for some 
 * sets, but all sets exceed the given threshold.
 * 
 * @author Marcin Skirzynski
 *
 */
public class PruneFalsePositives extends Operator {
	
	public static final String PARAMETER_MIN_SUPPORT = "min_support";
	
	private InputPort itemSetsInput = getInputPorts().createPort("frequent sets", FrequentItemSets.class);
	private InputPort exampleSetInput = getInputPorts().createPort("example set", ExampleSet.class);
	private OutputPort itemSetsOutput = getOutputPorts().createPort("frequent sets");
	private OutputPort exampleSetOutput = getOutputPorts().createPort("example set");

	/**
	 * Creates this operator and the transformer rules.
	 */
	public PruneFalsePositives(OperatorDescription description) {
		super(description);
		getTransformer().addPassThroughRule(itemSetsInput, itemSetsOutput);
		getTransformer().addPassThroughRule(exampleSetInput, exampleSetOutput);
		exampleSetInput.addPrecondition(new ExampleSetPrecondition(exampleSetInput, Ontology.BINOMINAL));
	}

	/**
	 * Updates the frequencies
	 */
	@Override
	public void doWork() throws OperatorException {
		ExampleSet exampleSet = exampleSetInput.getData();
		FrequentItemSets itemSets = itemSetsInput.getData();
		

		// indices for the positive values of the attributes
		double[] positiveIndices = getPositiveIndices(exampleSet.getAttributes());
		
		int n = exampleSet.size();
		double minSupp = getParameterAsDouble(PARAMETER_MIN_SUPPORT);
		
		FrequentItemSets result = new FrequentItemSets(n);
		Map<FrequentItemSet, Integer> check = new HashMap<FrequentItemSet, Integer>();
		
		int minFrequency = (int) (minSupp * n);

		for(FrequentItemSet itemSet : itemSets) {

			/*
			 * The given frequencies are minimum frequencies, so if they already
			 * exceed the threshold than it can be returned as frequent although
			 * the frequency might be not correct.
			 * 
			 * For not already frequent itemsets the correct frequency has to be computed
			 * to check if they are frequent anyway.
			 */
			if(itemSet.getFrequency() >= minFrequency) {
				result.addFrequentSet(itemSet);
			} else {
				check.put(itemSet, 0);
			}
		}
		
		int examplesPerDot = 100;
		System.err.println("Checking " + check.size() + " itemsets.");
		System.err.println("Iterating over " + exampleSet.size() + " examples - each dot are " + examplesPerDot);
		int examples = 0;
		for(Example example : exampleSet) {
			// The items from this example as strings
			Set<String> items = getItemsAsString(example, positiveIndices);
			
			// Mark all sets which exceed the threshold this time
			Collection<FrequentItemSet> exceeded = new ArrayList<FrequentItemSet>();
			for(FrequentItemSet itemSet : check.keySet()) {
				if(supports(items, itemSet)) {
					int freq = check.get(itemSet);
					if((freq + 1) > minFrequency)
						exceeded.add(itemSet);
					else
						check.put(itemSet, freq + 1);
				}
			}
			
			for(FrequentItemSet itemSet : exceeded) {
				check.remove(itemSet);
				result.addFrequentSet(new FrequentItemSet(new ArrayList<Item>(itemSet.getItems()), minFrequency));
			}
			
			checkForStop();
			examples++;
			if(examples % examplesPerDot == 0)
				System.err.print('.');
			
		}
		System.err.println('.');
		
		itemSetsOutput.deliver(result);
		exampleSetOutput.deliver(exampleSet);
	}
	
	public Set<String> getItemsAsString(Example example, double[] positiveIndices) {
		Set<String> result = new HashSet<String>();
		int i = 0;
		for(Attribute attribute : example.getAttributes()) {
			double value = example.getValue(attribute);
			if(value==positiveIndices[i++])
				result.add(attribute.getName());
		}
		return result;
	}
	
	public boolean supports(Set<String> items, FrequentItemSet itemSet) {
		for(Item item : itemSet.getItems()) {
			if(!items.contains(item.toString()))
				return false;
		}
		return true;
	}
	
	/**
	 * Get a cache for the positive indices for faster access.
	 */
	private double[] getPositiveIndices(Attributes attributes) {
		double[] positiveIndices = new double[attributes.size()];
		int i = 0;
		for (Attribute attribute : attributes) {
			positiveIndices[i++] = attribute.getMapping().getPositiveIndex();
		}
		return positiveIndices;
	}

	
	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.add(new ParameterTypeDouble(PARAMETER_MIN_SUPPORT, "The minimal support necessary in order to be added as a frequent item (set).", 0.0d, 1.0d, 0.0d));
		return types;
	}

}
