package com.rapidminer.operator.preprocessing.discretization;

import java.util.ArrayList;
import java.util.List;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.Statistics;
import com.rapidminer.example.set.SortedExampleSet;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.UserError;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeDouble;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.conditions.BooleanParameterCondition;

/**
 * Implementation of the discretization step used in the MAFIA-algorithm.
 *  
 * @author Marcin Skirzynski
 */
public class MafiaDiscretization extends MapReduceDiscretization {

	static {
		registerDiscretizationOperator(MafiaDiscretization.class);
	}

	public static final String PARAMETER_BETA = "beta";
	
	/** Indicates the number of used bins. */
	public static final String PARAMETER_INITIAL_NUMBER_OF_BINS = "initial_number_of_bins";
	
	/** Indicates the number of used bins. */
	public static final String PARAMETER_NUMBER_OF_BINS = "number_of_bins";
	
	public static final String PARAMETER_DEFINE_BOUNDARIES = "define_boundaries";
	
	public static final String PARAMETER_MIN_VALUE = "min_value";

	public static final String PARAMETER_MAX_VALUE = "max_value";

	public MafiaDiscretization(OperatorDescription description) {
		super(description);
	}

	/**
	 * Returns an array of cutpoints which will be used for the intervals.
	 */
	public double[] discretize(Attribute attribute, ExampleSet exampleSet) throws UserError {

		double beta = getParameterAsDouble(PARAMETER_BETA);
		double[] ranges = getInitialBinning(exampleSet, attribute);
		
		if(ranges.length<=3)
			return ranges;

		// Counts for the histogram 
		int[] counts = new int[ranges.length];
		
		ExampleSet sorted = new SortedExampleSet(exampleSet, attribute, SortedExampleSet.INCREASING);
		int j = 0;
		// to ranges.length-2 to not deal with positive infty.
		for(int i=0; i<ranges.length-2 && j<sorted.size(); i++) {
			double left = ranges[i];
			double right = ranges[i+1];
			
			// Counting how many examples are within the interval
			double currentExampleVal = sorted.getExample(j).getNumericalValue(attribute);
			while(currentExampleVal>=left && currentExampleVal<right) {
				counts[i] = counts[i] + 1;
				if(++j < sorted.size())
					currentExampleVal = sorted.getExample(j).getNumericalValue(attribute);
				else
					break;
			}
		}
		
		List<Double> mergedBins = new ArrayList<Double>();
		mergedBins.add(ranges[0]);
		for(int i=0; i<ranges.length-1;) {
			int k = 1;
			int c1 = counts[i];
			int c2 = counts[i+k];
			
			double fraction = getFraction(c1, c2);
			while(fraction<beta && (i+k+1)<counts.length) {
				k++;
				c1 = c2;
				c2 = counts[i+k];
				fraction = getFraction(c1, c2);
			}
			i += k;
			mergedBins.add(ranges[i]);
		}
		
		double[] result = new double[mergedBins.size()];
		for(int i=0; i<result.length; i++)
			result[i] = mergedBins.get(i);
		
		return result;
	}
	
	/**
	 * The fraction of the frequency which will be used with the beta
	 * parameter for deciding if to add an cutpoint or not. 
	 */
	private double getFraction(int bin1, int bin2) {
			int max = bin1 > bin2 ? bin1 : bin2;
			int min = bin1 <= bin2 ? bin1 : bin2;
			if(max==0)
				return 0.0;
			return 1 - (min / (double)max);
	}
	
	/**
	 * Creates an initial binning in equal width intervals. See {@link BinDiscretization}. 
	 */
	private double[] getInitialBinning( ExampleSet exampleSet, Attribute attribute) throws UserError {
		
		int numberOfBins = getParameterAsInt(PARAMETER_INITIAL_NUMBER_OF_BINS);
		double[] binRange = null;
		if (getParameterAsBoolean(PARAMETER_DEFINE_BOUNDARIES)) {
			double min = getParameterAsDouble(PARAMETER_MIN_VALUE);
			double max = getParameterAsDouble(PARAMETER_MAX_VALUE);
			if (min > max) {
				throw new UserError(this, 116, PARAMETER_MIN_VALUE + " and " + PARAMETER_MAX_VALUE, "minimum must be less than maximum");
			}
			binRange = new double[numberOfBins + 2];
			binRange[0] = min;
			for (int b = 1; b < numberOfBins; b++) {
				binRange[b] = min + (((double) b / (double) numberOfBins) * (max - min));
			}
			binRange[numberOfBins] = max;
			binRange[numberOfBins + 1] = Double.POSITIVE_INFINITY;
		} else {
			binRange = new double[numberOfBins];
			double min = exampleSet.getStatistics(attribute, Statistics.MINIMUM);
			double max = exampleSet.getStatistics(attribute, Statistics.MAXIMUM);
			for (int b = 0; b < numberOfBins - 1; b++) {
				binRange[b] = b==0 ? min : min + (((double) (b + 1) / (double) numberOfBins) * (max - min));
			}
			binRange[numberOfBins - 1] = Double.POSITIVE_INFINITY;
		}
		return binRange;
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();

		
		ParameterType type = new ParameterTypeInt(PARAMETER_INITIAL_NUMBER_OF_BINS, "Defines the initial number of bins which will be merged.", 2, Integer.MAX_VALUE, 15);
		type.setExpert(false);
		types.add(type);
		
		type = new ParameterTypeDouble(PARAMETER_BETA, "The threshold which determines if two bins will be merged", 0, 1, 0.3);
		type.setExpert(false);
		types.add(type);
		
		type = new ParameterTypeInt(PARAMETER_NUMBER_OF_BINS, "Defines the number of bins which should be used for each attribute if there is only one bin left.", 2, Integer.MAX_VALUE, 5);
		type.setExpert(false);
		types.add(type);
		
		type = new ParameterTypeBoolean(PARAMETER_DEFINE_BOUNDARIES, "Define the boundraries for the bin calculation.", false);
		types.add(type);
		type = new ParameterTypeDouble(PARAMETER_MIN_VALUE, "The minimum value for the binning range.", Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, true);
		type.registerDependencyCondition(new BooleanParameterCondition(this, PARAMETER_DEFINE_BOUNDARIES, true, true));
		types.add(type);
		type = new ParameterTypeDouble(PARAMETER_MAX_VALUE, "The maximum value for the binning range.", Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, true);
		type.registerDependencyCondition(new BooleanParameterCondition(this, PARAMETER_DEFINE_BOUNDARIES, true, true));
		types.add(type);

		return types;
	}
	
}
