package com.rapidminer.operator.preprocessing.discretization;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.mapreduce.MapReduceException;
import com.rapidminer.mapreduce.MapReduceExecutor;
import com.rapidminer.mapreduce.MapReduceExecutorLocal;
import com.rapidminer.mapreduce.Mapper;
import com.rapidminer.mapreduce.MapperFactory;
import com.rapidminer.mapreduce.Reducer;
import com.rapidminer.mapreduce.ReducerFactory;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.Value;
import com.rapidminer.operator.preprocessing.PreprocessingModel;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeCategory;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.conditions.BooleanParameterCondition;
import com.rapidminer.parameter.conditions.EqualTypeCondition;
import com.rapidminer.tools.container.Pair;

/**
 * Abstract implementation for a discretization operator which can be used
 * with the map-reduce framework in parallel.
 * 
 * @author Marcin Skirzynski
 */
public abstract class MapReduceDiscretization 
				extends 
					AbstractDiscretizationOperator
				implements 	
					MapperFactory<Pair<Attribute, ExampleSet>, Pair<Attribute, double[]>>, 
					ReducerFactory<Pair<Attribute,double[]>, Map<Attribute, double[]>> {
						
	private static final String PARAMETER_THREADS = "threads";
						
	private static final String RANGE_NAME_BASE = "range";
	
	/** Indicates if long range names should be used. */
	public static final String PARAMETER_RANGE_NAME_TYPE = "range_name_type";

	public static final String PARAMETER_AUTOMATIC_NUMBER_OF_DIGITS = "automatic_number_of_digits";

	public static final String PARAMETER_NUMBER_OF_DIGITS = "number_of_digits";
	
	private int numberOfIntervals = 0;

	/**
	 * Adds a value for logging which provides the number of intervals created.
	 */
	public MapReduceDiscretization(OperatorDescription description) {
		super(description);
		addValue(new Value("intervals", "Number of intervals created") {
			@Override
			public boolean isNominal() {
				return false;
			}
			@Override
			public Object getValue() {
				return (double)numberOfIntervals;
			}
		});
	}

	/**
	 * Creates a model by starting the map-reduce framework and invoke the abstract
	 * method {@link MapReduceDiscretization#discretize(Attribute, ExampleSet)}
	 * for every attribute.
	 */
	@Override
	public PreprocessingModel createPreprocessingModel(ExampleSet exampleSet)
			throws OperatorException {
		DiscretizationModel model = new DiscretizationModel(exampleSet);
		
		exampleSet.recalculateAllAttributeStatistics();
		numberOfIntervals = 0;
		
		int threads = getParameterAsInt(PARAMETER_THREADS);
		MapReduceExecutor<Pair<Attribute, ExampleSet>, Pair<Attribute, double[]>, Map<Attribute, double[]>> mr =
				new MapReduceExecutorLocal<Pair<Attribute, ExampleSet>, Pair<Attribute, double[]>, Map<Attribute,double[]>>(threads);

		Collection<Pair<Attribute,ExampleSet>> chunks = new ArrayList<Pair<Attribute,ExampleSet>>();
		for(Attribute attribute : exampleSet.getAttributes()) {
			if(attribute.isNumerical()) // Skip not numerical attributes
				chunks.add(new Pair<Attribute, ExampleSet>(attribute, exampleSet));
		}
		
		Map<Attribute, double[]> ranges;
		try {
			ranges = mr.mapReduce(chunks, this, this);
		} catch (MapReduceException e) {
			Throwable caught = e.getCause();
			if(caught instanceof UserError)
				throw (UserError)caught;
			throw new RuntimeException(caught);
		}
		
		for(double[] cuts : ranges.values()) {
			numberOfIntervals += (cuts.length+2);
		}
		
		// determine number of digits
		int numberOfDigits = -1;
		if (getParameterAsBoolean(PARAMETER_AUTOMATIC_NUMBER_OF_DIGITS) == false) {
			numberOfDigits = getParameterAsInt(PARAMETER_NUMBER_OF_DIGITS);
		}

		model.setRanges(ranges, RANGE_NAME_BASE, getParameterAsInt(PARAMETER_RANGE_NAME_TYPE), numberOfDigits);
		return model;
	}

	/**
	 * Every implementation shall implement this method for discretization of a single attribute. 
	 */
	public abstract double[] discretize(Attribute attribute, ExampleSet exampleSet) throws UserError;

	@Override
	public Class<? extends PreprocessingModel> getPreprocessingModelClass() {
		return DiscretizationModel.class;
	}
	
	/**
	 * Create the map-function objects used by the map-reduce framework.
	 */
	@Override
	public Mapper<Pair<Attribute,ExampleSet>, Pair<Attribute, double[]>> createMapper() {
		return new Mapper<Pair<Attribute,ExampleSet>, Pair<Attribute,double[]>>() {
			@Override
			public Pair<Attribute, double[]> map(
					Pair<Attribute, ExampleSet> input)
					throws MapReduceException {
				double[] discretization;
				try {
					discretization = discretize(input.getFirst(), input.getSecond());
					return new Pair<Attribute, double[]>(input.getFirst(), discretization);
				} catch (UserError e) {
					throw new MapReduceException(e);
				}
			}
		};
	}
	
	/**
	 * Create the reduce-function objects used by the map-reduce framework.
	 */
	@Override
	public Reducer<Pair<Attribute, double[]>, Map<Attribute, double[]>> createReducer() {
		return new Reducer<Pair<Attribute,double[]>, Map<Attribute,double[]>>() {
			@Override
			public Map<Attribute, double[]> reduce(
					Collection<Pair<Attribute, double[]>> input)
					throws MapReduceException {
				Map<Attribute, double[]> result = new HashMap<Attribute, double[]>();
				for(Pair<Attribute, double[]> pair : input) {
					result.put(pair.getFirst(), pair.getSecond());
				}
				return result;
			}
		};
	}
	
	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		
		ParameterType type = new ParameterTypeInt(PARAMETER_THREADS, "Number of threads to compute", 1, Integer.MAX_VALUE, Runtime.getRuntime().availableProcessors());
		type.setExpert(true);
		types.add(type);

		types.add(new ParameterTypeCategory(PARAMETER_RANGE_NAME_TYPE, "Indicates if long range names including the limits should be used.", DiscretizationModel.RANGE_NAME_TYPES, DiscretizationModel.RANGE_NAME_LONG));		

		type = new ParameterTypeBoolean(PARAMETER_AUTOMATIC_NUMBER_OF_DIGITS, "Indicates if the number of digits should be automatically determined for the range names.", true);
		type.registerDependencyCondition(new EqualTypeCondition(this, PARAMETER_RANGE_NAME_TYPE, DiscretizationModel.RANGE_NAME_TYPES, false, DiscretizationModel.RANGE_NAME_INTERVAL));
		types.add(type);

		type = new ParameterTypeInt(PARAMETER_NUMBER_OF_DIGITS, "The minimum number of digits used for the interval names.", 1, Integer.MAX_VALUE, 3);
		type.registerDependencyCondition(new BooleanParameterCondition(this, PARAMETER_AUTOMATIC_NUMBER_OF_DIGITS, false, false));
		types.add(type);

		return types;
	}

	

}
