package com.rapidminer.operator.preprocessing.discretization;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.set.SortedExampleSet;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.UserError;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeDouble;

/**
 * The implementation of the neighborhood-method for the discretization of 
 * the data space.
 * 
 * @author Marcin Skirzynski
 */
public class NeighborhoodDiscretization extends MapReduceDiscretization {
	
	private static final String PARAMETER_EPSILON = "epsilon";
	private static final String PARAMETER_BETA = "beta";
	private static final String PARAMETER_MIN_SUPPORT = "min_supp";

	static {
		registerDiscretizationOperator(NeighborhoodDiscretization.class);
	}
	
	public NeighborhoodDiscretization(OperatorDescription description) {
		super(description);
	}

	/**
	 * Returns an array of cutpoints which will be used for the intervals.
	 */
	@Override
	public double[] discretize(Attribute attribute, ExampleSet exampleSet) throws UserError {
		List<Double> cuts = new ArrayList<Double>();
		SortedExampleSet sorted = new SortedExampleSet(exampleSet, attribute, SortedExampleSet.INCREASING);
		
		double beta = getParameterAsDouble(PARAMETER_BETA);
		double epsilon = getParameterAsDouble(PARAMETER_EPSILON);
		double minSupp = getParameterAsDouble(PARAMETER_MIN_SUPPORT);
	
		Queue<Double> left = new LinkedList<Double>();
		Queue<Double> right = new LinkedList<Double>();
		
		Double currentValue = null;
		int i = 0;
		right.add(sorted.getExample(i).getNumericalValue(attribute));
		
		while((currentValue=right.poll())!=null || i<sorted.size()) {
			if(currentValue==null) {
				currentValue = sorted.getExample(++i).getNumericalValue(attribute);
			}
			double minValue = currentValue - epsilon;
			double maxValue = currentValue + epsilon;

			// Update left queue
			while(!left.isEmpty() && left.peek()<minValue) {
				left.poll();
			}
			
			boolean stop = false;
			while(!stop && ++i<sorted.size()) {
				double lookaheadValue = sorted.getExample(i).getNumericalValue(attribute);
				if(lookaheadValue<maxValue) {
					right.add(lookaheadValue);
				} else {
					i--;
					stop = true;
				}
			}
			
			// Determine result
			int lcount = left.size();
			int rcount = right.size();
			double support = (lcount + rcount)/(double)exampleSet.size();
			
			if(support>minSupp) {
				double fraction = getFraction(rcount, lcount);
				if(fraction>beta) {
					cuts.add(currentValue);
					left = right;
					right = new LinkedList<Double>();
					if(i<sorted.size())
						right.add(sorted.getExample(i).getNumericalValue(attribute));
				} else {
					left.add(currentValue);
				}
			} else {
				left.add(currentValue);
			}
		}
		
		/*
		 *  If there are any points left after the last cut
		 *  the last point will be inserted as cut.
		 */
		if(!left.isEmpty()) {
			double last = cuts.get(cuts.size()-1);
			double max = Collections.max(left);
			if(max>=last)
				cuts.add(max);
		}
		
		
		double[] result = new double[cuts.size()];
		int j=0;
		for(double cut : cuts) {
			result[j++] = cut;
		}
		return result;
	}

	
	/**
	 * The fraction of the frequency which will be used with the beta
	 * parameter for deciding if to add an cutpoint or not. 
	 */
	private static double getFraction(int bin1, int bin2) {
		int max = Math.max(bin1, bin2);
		if(max==0) 
			return 0.0;
		return Math.abs(bin1 - bin2)/(double)max;
	}
	
	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();

		types.add(new ParameterTypeDouble(PARAMETER_EPSILON, "The epsilon for the neighborhood for which the frequency will be measured", 0.0, Double.MAX_VALUE, 1.0));
		types.add(new ParameterTypeDouble(PARAMETER_BETA, "The maximal percental difference of frequency of the left and right neighborhood before this example will be used as cutpoint", 0.0, 1.0, 0.3));
		types.add(new ParameterTypeDouble(PARAMETER_MIN_SUPPORT, "The minimal support which determines that point are just noise", 0.0, 1.0, 0.01));
		
		return types;
	}

	
}
