package com.rapidminer.operator.preprocessing.statistics;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.subspace.MinimalDescription;
import com.rapidminer.subspace.MinimalDescriptionCollection;

/**
 * THe contingency table used to calculate the jaccard- and rand-indices.
 *   
 * @author Marcin Skirzynski
 */
public class ContingencyTable {
	
	private int f00;
	
	private int f11;
	
	private int f10;
	
	private int f01;
	
	/**
	 * Creates the table for the given clusters and the data set. Neither the example set nor
	 * the minimal descriptions will be stored. 
	 */
	public ContingencyTable(ExampleSet exampleSet, MinimalDescriptionCollection descriptions) {
		Attribute label = exampleSet.getAttributes().getLabel();
		if(label==null) {
			throw new IllegalArgumentException("ExampleSet has no label set - can not comput contingency table");
		}
		
		for(Example e1 : exampleSet ) {
			for(Example e2 : exampleSet ) {
				boolean sameLabel = sameLabel(e1,e2);
				boolean sameCluster = sameCluster(e1, e2, descriptions);
				
				if(sameLabel && sameCluster) {
					f11++;
				} else if(sameLabel && !sameCluster) {
					f10++;
				} else if(!sameLabel && sameCluster) {
					f01++;
				} else {
					f00++;
				}
			}
		}
	}
	
	/**
	 * Returns the jaccard index from the table
	 */
	public double getJaccardIndex() {
		return f11 / (double) (f01 + f10 + f11);
	}
	
	/**
	 * Returns the rand index from the table
	 */
	public double getRandIndex() {
		return (f11 + f00) / (double) (f00 + f10 + f01 + f11);
	}
	
	/**
	 * Returns <code>true</code> if the examples have the same label. 
	 */
	public boolean sameLabel(Example e1, Example e2) {
		return e1.getLabel()==e2.getLabel();
	}
	
	/**
	 * Returns <code>true</code> if the examples are in the same cluster.
	 */
	public boolean sameCluster(Example e1, Example e2, MinimalDescriptionCollection descriptions) {
		for(MinimalDescription description : descriptions.getObjects()) {
			boolean e1Contained = description.contains(e1);
			boolean e2Contained = description.contains(e2);
			if(e1Contained && !e2Contained)
				return false;
			if(e2Contained && !e1Contained)
				return false;
			if(e1Contained && e2Contained)
				return true;
		}
		return false;
	}
	
	public int getf00() {
		return f00;
	}
	
	public int getf11() {
		return f11;
	}
	
	public int getf10() {
		return f10;
	}
	
	public int getf01() {
		return f01;
	}

}
