package com.rapidminer.operator.preprocessing.statistics;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.example.table.DataRow;
import com.rapidminer.example.table.DataRowFactory;
import com.rapidminer.example.table.MemoryExampleTable;
import com.rapidminer.operator.IOObjectCollection;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.AttributeMetaData;
import com.rapidminer.operator.ports.metadata.ExampleSetMetaData;
import com.rapidminer.operator.ports.metadata.GenerateNewExampleSetMDRule;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.subspace.MinimalDescription;
import com.rapidminer.subspace.MinimalDescriptionCollection;
import com.rapidminer.tools.Ontology;

/**
 * Computes several statistics about a given subspace cluster, inject this
 * into the minimal descriptions and returns an example set which contains
 * all statistics for all cluster.
 * 
 * @author Marcin Skirzynski
 */
public class SubspaceClusterStatistics extends Operator {
	
	private static final String DIMENSION_ATTR = "dimension";
	private static final String CLUSTERS_ATTR = "clusters";
	private static final String COVERAGE_ATTR = "coverage";
	private static final String RAND_ATTR = "rand-index";
	private static final String JACCARD_ATTR = "jaccard-index";
	private static final String SILHOUETTE_ATTR = "silhouette-coeffizient";

	private InputPort minimalDescriptionInput = getInputPorts().createPort("minimal description", IOObjectCollection.class);
	private InputPort exampleSetInput = getInputPorts().createPort("example set", ExampleSet.class);
	
	private OutputPort statisticsOutput = getOutputPorts().createPort("statistics");
	private OutputPort minimalDescriptionOutput = getOutputPorts().createPassThroughPort("minimal description");
	private OutputPort exampleSetOutput = getOutputPorts().createPassThroughPort("example set");
	
	public SubspaceClusterStatistics(OperatorDescription description) {
		super(description);
		getTransformer().addPassThroughRule(minimalDescriptionInput, minimalDescriptionOutput);
		getTransformer().addPassThroughRule(exampleSetInput, exampleSetOutput);
	
		getTransformer().addRule(new GenerateNewExampleSetMDRule(statisticsOutput) {
			@Override
			public MetaData modifyMetaData(ExampleSetMetaData unmodifiedMetaData) {
				unmodifiedMetaData.addAttribute(new AttributeMetaData(DIMENSION_ATTR, Ontology.INTEGER));
				unmodifiedMetaData.addAttribute(new AttributeMetaData(CLUSTERS_ATTR, Ontology.INTEGER));
				unmodifiedMetaData.addAttribute(new AttributeMetaData(COVERAGE_ATTR, Ontology.REAL));
				unmodifiedMetaData.addAttribute(new AttributeMetaData(RAND_ATTR, Ontology.REAL));
				unmodifiedMetaData.addAttribute(new AttributeMetaData(JACCARD_ATTR, Ontology.REAL));
				unmodifiedMetaData.addAttribute(new AttributeMetaData(SILHOUETTE_ATTR, Ontology.REAL));
				return unmodifiedMetaData;
			}
		});
	}
	
	@Override
	public void doWork() throws OperatorException {
		ExampleSet exampleSet = exampleSetInput.getData(ExampleSet.class);
		@SuppressWarnings("unchecked")
		IOObjectCollection<MinimalDescriptionCollection> md = minimalDescriptionInput.getData(IOObjectCollection.class);
		
		Attribute label = exampleSet.getAttributes().getLabel();
		
		for(Example example : exampleSet) {
			for(MinimalDescriptionCollection mds : md.getObjects()) {
				for(MinimalDescription description : mds.getObjects()) {
					if(description.contains(example)) {
						if(label==null) {
							description.addContainedExample();
						} else {
							String labelName = label.getName();
							labelName += "=" + example.getValueAsString(label);
							description.addContainedExample(labelName);
						}
					}
				}
			}
		}
		
		Attribute dimensionAttr = AttributeFactory.createAttribute(DIMENSION_ATTR, Ontology.INTEGER);
		Attribute clusterAttr = AttributeFactory.createAttribute(CLUSTERS_ATTR, Ontology.INTEGER);
		Attribute coverageAttr = AttributeFactory.createAttribute(COVERAGE_ATTR, Ontology.REAL);
		Attribute randAttr = AttributeFactory.createAttribute(RAND_ATTR, Ontology.REAL);
		Attribute jaccardAttr = AttributeFactory.createAttribute(JACCARD_ATTR, Ontology.REAL);
		Attribute silhouetteAttr = AttributeFactory.createAttribute(SILHOUETTE_ATTR, Ontology.REAL);
		
		Attribute[] attributes = new Attribute[] {dimensionAttr, clusterAttr, coverageAttr, randAttr, jaccardAttr, silhouetteAttr};

		MemoryExampleTable table = new MemoryExampleTable(attributes);
		DataRowFactory dataRowFactory = new DataRowFactory(DataRowFactory.TYPE_DOUBLE_ARRAY, '.');

		for(MinimalDescriptionCollection mds : md.getObjects()) {
			
			int absoluteCoverage = 0;
			for(MinimalDescription description : mds.getObjects()) {
				absoluteCoverage += description.getAbsoluteCoverage();
			}
			
			int dimension = mds.getDimension();
			int clusters = mds.size();
			double coverage = absoluteCoverage/(double)exampleSet.size();
			
			ContingencyTable contigencyTable = new ContingencyTable(exampleSet, mds);
			double rand = contigencyTable.getRandIndex();
			double jaccard = contigencyTable.getJaccardIndex();
			
			mds.setCoverage(coverage);
			mds.setJaccard(jaccard);
			mds.setRand(rand);
			mds.setExampleSetSize(exampleSet.size());
			mds.calculateSilhouette(exampleSet);
			
			double silhouette = mds.getAvgSilhouette();
			
			DataRow row = dataRowFactory.create(new Double[] {(double)dimension, (double)clusters, coverage, rand, jaccard, silhouette}, attributes);

			table.addDataRow(row);
		}
	
		minimalDescriptionOutput.deliver(md);
		exampleSetOutput.deliver(exampleSet);
		statisticsOutput.deliver(table.createExampleSet());
	}

}
