package com.rapidminer.operator.preprocessing.transformation;

import java.util.ArrayList;
import java.util.List;

import com.rapidminer.operator.IOObjectCollection;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeDouble;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.subspace.MinimalDescription;
import com.rapidminer.subspace.MinimalDescriptionCollection;

/**
 * Filters descriptions by several indices.
 * 
 * @author Marcin Skirzynski
 */
public class FilterMinimalDescriptions extends Operator {
	
	private static final String PARAMETER_MIN_CLUSTER_COUNT = "min_cluster_count";
	
	private static final String PARAMETER_MIN_DIMENSION = "min_dimension";
	
	private static final String PARAMETER_MIN_JACCARD = "min_jaccard";
	
	private static final String PARAMETER_MIN_RAND = "min_rand";

	private InputPort minimalDescription = getInputPorts().createPort("minimal description", IOObjectCollection.class);
	
	private OutputPort prunedMinimalDescription = getOutputPorts().createPort("pruned");
	
	private OutputPort passthroughPort = getOutputPorts().createPassThroughPort("original");
	
	public FilterMinimalDescriptions(OperatorDescription description) {
		super(description);
		getTransformer().addPassThroughRule(minimalDescription, passthroughPort);
	}
	
	@Override
	public void doWork() throws OperatorException {
		IOObjectCollection<MinimalDescriptionCollection> input = minimalDescription.getData();
		IOObjectCollection<MinimalDescriptionCollection> result = new IOObjectCollection<MinimalDescriptionCollection>();
		
		int minClusterCount = getParameterAsInt(PARAMETER_MIN_CLUSTER_COUNT);
		int minDimension = getParameterAsInt(PARAMETER_MIN_DIMENSION);
		double minJaccard = getParameterAsDouble(PARAMETER_MIN_JACCARD);
		double minRand = getParameterAsDouble(PARAMETER_MIN_RAND);
		
		for(MinimalDescriptionCollection collection : input.getObjects()) {
			List<MinimalDescription> notPruned = new ArrayList<MinimalDescription>(collection.getObjects());
			
			if(minClusterCount>=0) {
				if(collection.size()<minClusterCount)
					notPruned.removeAll(collection.getObjects());
			}
			if(minDimension>=0) {
				if(collection.getDimension()<minDimension)
					notPruned.removeAll(collection.getObjects());
			}
			if(minJaccard>=0.0) {
				if(collection.getJaccard()<minJaccard)
					notPruned.removeAll(collection.getObjects());
			}
			if(minRand>=0.0) {
				if(collection.getRand()<minRand)
					notPruned.removeAll(collection.getObjects());
			}
			
			if(!notPruned.isEmpty()) {
				MinimalDescriptionCollection npColl = new MinimalDescriptionCollection(notPruned, collection.getDimension());
				npColl.setCoverage(collection.getCoverage());
				npColl.setJaccard(collection.getJaccard());
				npColl.setRand(collection.getRand());
				result.add(npColl);
				
			}
		}

		prunedMinimalDescription.deliver(result);
		passthroughPort.deliver(input);
	}
	
	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		
		ParameterType type = new ParameterTypeInt(PARAMETER_MIN_CLUSTER_COUNT, "Only subspace with at least the given cluster count will be used; not used if -1", -1, Integer.MAX_VALUE, -1);
		types.add(type);
		type = new ParameterTypeInt(PARAMETER_MIN_DIMENSION, "Only subspace with at least the given dimension will be used; not used if -1", -1, Integer.MAX_VALUE, -1);
		types.add(type);
		type = new ParameterTypeDouble(PARAMETER_MIN_JACCARD, "Only subspace with at least the given jaccard index will be used; not used if -1", -1, 1.0, -1);
		types.add(type);
		type = new ParameterTypeDouble(PARAMETER_MIN_RAND, "Only subspace with at least the given rand index will be used; not used if -1", -1, 1.0, -1);
		types.add(type);
		
		return types;
	}

}
