package com.rapidminer.operator.preprocessing.transformation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.rapidminer.operator.IOObjectCollection;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.subspace.Interval.Direction;
import com.rapidminer.subspace.MinimalDescription;
import com.rapidminer.subspace.MinimalDescriptionCollection;
import com.rapidminer.subspace.Region;
import com.rapidminer.subspace.Subspace;
import com.rapidminer.subspace.SubspaceCluster;
import com.rapidminer.subspace.SubspaceClustering;
import com.rapidminer.subspace.Unit;
import com.rapidminer.tools.RandomGenerator;

/**
 * Takes a subspace clustering and finds the minimal description for it.
 * 
 * @author Marcin Skirzynski
 */
public class FindMinimalDescription extends Operator {
	
	private InputPort subspaceClusteringInput = getInputPorts().createPort("subspace clustering", SubspaceClustering.class);
	
	private OutputPort minimalDescriptionsOutput = getOutputPorts().createPort("minimal descriptions");
	
	private OutputPort subspaceClusteringOutput = getOutputPorts().createPassThroughPort("subspace clustering");
	

	public FindMinimalDescription(OperatorDescription description) {
		super(description);
		getTransformer().addPassThroughRule(subspaceClusteringInput, subspaceClusteringOutput);
		getTransformer().addGenerationRule(minimalDescriptionsOutput, IOObjectCollection.class);
	}
	
	@Override
	public void doWork() throws OperatorException {
		SubspaceClustering subspaceClustering = subspaceClusteringInput.getData();
		RandomGenerator random = RandomGenerator.getRandomGenerator(this);

		Map<Subspace, List<MinimalDescription>> descriptions = new HashMap<Subspace, List<MinimalDescription>>();
		for(SubspaceCluster cluster : subspaceClustering.getClusters()) {
			Set<Region> regions = computeRegions(cluster, random);
			if(!descriptions.containsKey(cluster.getSubspace()))
				descriptions.put(cluster.getSubspace(), new ArrayList<MinimalDescription>());
			descriptions.get(cluster.getSubspace()).add(new MinimalDescription(regions));
		}
		
		subspaceClusteringOutput.deliver(subspaceClustering);
		
		IOObjectCollection<IOObjectCollection<MinimalDescription>> result = 
				new IOObjectCollection<IOObjectCollection<MinimalDescription>>();
		for(final Subspace subspace : descriptions.keySet()) {
			result.add(new MinimalDescriptionCollection(descriptions.get(subspace), subspace.size()));
		}
		minimalDescriptionsOutput.deliver(result);
	}
	
	/**
	 * Creates all regions which contain the cluster. 
	 */
	public Set<Region> computeRegions(SubspaceCluster cluster, RandomGenerator random) {
		Set<Unit> units = cluster.getUnits();
		if(units.isEmpty())
			throw new IllegalArgumentException("Cluster consist of no units");

		Set<Region> result = new HashSet<Region>();

		// The simple case ... if a cluster is just one unit so the intervals from this units are the region.
		if(units.size()==1) {
			Region region = Region.createRegion(cluster.getSubspace(), units);
			result.add(region);
			return result;
		}

		Set<Unit> notInRegion = new HashSet<Unit>(units);
		
		while(!notInRegion.isEmpty()) {
			Unit unit = selectRandomUnit(notInRegion, random);	
			
			Set<Unit> expanded = new HashSet<Unit>();
			expanded.add(unit);
			expand(expanded, cluster);
			notInRegion.removeAll(expanded);
			result.add(Region.createRegion(cluster.getSubspace(), expanded));
		}
		
		
		return result;
	}
	
	/**
	 * Expand the units for all attributes of the cluster
	 */
	public void expand(Set<Unit> toExpand, SubspaceCluster cluster) {
		for( String attribute : cluster.getSubspace().getAttributeNames() ) {
			expand(toExpand, cluster, attribute);
		}
	}
	
	/**
	 * Expands to the left and right for the given attribute.
	 */
	public void expand(Set<Unit> toExpand, SubspaceCluster cluster, String attribute) {
		while(expand(toExpand, cluster, attribute, Direction.LEFT)) {}
		while(expand(toExpand, cluster, attribute, Direction.RIGHT)) {}
	}

	/**
	 * Expands the units in a single direction.
	 */
	public boolean expand(Set<Unit> toExpand, SubspaceCluster cluster, String attribute, Direction direction ) {
		Set<Unit> expanded = new HashSet<Unit>();
		for(Unit unit : toExpand) {
			Unit neighbor = cluster.getNeighbor(unit, direction, attribute);
			if(neighbor==null)
				return false;
			expanded.add(neighbor);
		}
		toExpand.addAll(expanded);
		return true;
	}
	

	/**
	 * Selects a random unit from the set of units.
	 */
	public Unit selectRandomUnit(Set<Unit> units, RandomGenerator random) {
		if(units.size()==1)
			return units.iterator().next();
		return new ArrayList<Unit>(units).get(random.nextInt(units.size()));
	}

}
