package com.rapidminer.operator.preprocessing.transformation;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.learner.associations.FrequentItemSet;
import com.rapidminer.operator.learner.associations.FrequentItemSets;
import com.rapidminer.operator.learner.associations.Item;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.subspace.ClusterSearch;
import com.rapidminer.subspace.Interval;
import com.rapidminer.subspace.ItemParser;
import com.rapidminer.subspace.Subspace;
import com.rapidminer.subspace.SubspaceCluster;
import com.rapidminer.subspace.SubspaceClustering;
import com.rapidminer.subspace.Unit;

/**
 * Transforms a frequent item set to a subspace cluster.
 * 
 * @author Marcin Skirzynski
 *
 */
public class FrequentItemset2SubspaceCluster extends Operator {

	private InputPort fisInput = getInputPorts().createPort("frequentItemSets", FrequentItemSets.class);
	
	private OutputPort subspaceClusteringOutput = getOutputPorts().createPort("subspace clustering");
	
	public FrequentItemset2SubspaceCluster(OperatorDescription description) {
		super(description);
		
		getTransformer().addGenerationRule(subspaceClusteringOutput, SubspaceClustering.class);
	}
	
	@Override
	public void doWork() throws OperatorException {
		FrequentItemSets fis = fisInput.getData(FrequentItemSets.class);
		
		ItemParser parser = new ItemParser();
		
		/*
		 * A map which maps a subspace (i.e. a set of attribute names)
		 * to the dense units within it. 
		 */
		Map<Subspace,Set<Unit>> units = new HashMap<Subspace, Set<Unit>>();
		for(FrequentItemSet set : fis) {
			if(set.getNumberOfItems()==0)
				continue;
			Collection<Interval> intervals = new ArrayList<Interval>();
			Collection<String> booleanAttributes = new ArrayList<String>();
			for(Item item : set.getItems()) {
				if(parser.isBoolean(item)) {
					booleanAttributes.add(parser.getBooleanAttributeName(item));
					continue;
				}
				try {
					Interval interval = parser.parse(item);
					double intervalDensity = (double)item.getFrequency()/(double)fis.getNumberOfTransactions(); 
					interval.setDensity(intervalDensity);
					intervals.add(interval);
				} catch (ParseException e) {
					throw new UserError(this, e, 0);
				}
			}
			
			double unitDensity = (double)set.getFrequency()/(double)fis.getNumberOfTransactions();
			Unit unit = new Unit(intervals, unitDensity);
			unit.setBooleanAttributes(booleanAttributes);
			Subspace subspace = unit.getSubspace();
			
			if(!units.containsKey(subspace))
				units.put(subspace, new HashSet<Unit>());
			units.get(subspace).add(unit);
		}
		
		ClusterSearch search = new ClusterSearch();
		Collection<SubspaceCluster> subspaceCluster = new ArrayList<SubspaceCluster>();
		
		for( Subspace subspace : units.keySet() ) {
			Set<Unit> u = units.get(subspace);
			subspaceCluster.addAll(search.find(u));
		}
		
		subspaceClusteringOutput.deliver(new SubspaceClustering(subspaceCluster));
	}
	
}
