package com.rapidminer.operator.preprocessing.transformation;

import java.util.ArrayList;
import java.util.List;

import com.rapidminer.operator.IOObjectCollection;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeDouble;
import com.rapidminer.parameter.ParameterTypeString;
import com.rapidminer.subspace.MinimalDescription;
import com.rapidminer.subspace.MinimalDescriptionCollection;

/**
 * Pruned all minimal descriptions which label distribution varies
 * from a expected percentage by a given epsilon.
 * 
 * @author Marcin Skirzynski
 *
 */
public class PruneLabelDistribution extends Operator {
	
	private static final String PARAMETER_LABEL = "label";
	
	private static final String PARAMETER_PERCENTAGE = "expected_percentage";
	
	private static final String PARAMETER_EPSILON = "epsilon";

	private InputPort minimalDescription = getInputPorts().createPort("minimal description", IOObjectCollection.class);
	
	private OutputPort prunedMinimalDescription = getOutputPorts().createPort("pruned");
	
	private OutputPort passthroughPort = getOutputPorts().createPassThroughPort("original");
	
	public PruneLabelDistribution(OperatorDescription description) {
		super(description);
		getTransformer().addPassThroughRule(minimalDescription, passthroughPort);
	}
	
	@Override
	public void doWork() throws OperatorException {
		IOObjectCollection<MinimalDescriptionCollection> input = minimalDescription.getData();
		
		String label = getParameter(PARAMETER_LABEL);
		double expected = getParameterAsDouble(PARAMETER_PERCENTAGE);
		double epsilon = getParameterAsDouble(PARAMETER_EPSILON);
		
		IOObjectCollection<MinimalDescriptionCollection> result = new IOObjectCollection<MinimalDescriptionCollection>();
		
		for(MinimalDescriptionCollection collection : input.getObjects()) {
			
			List<MinimalDescription> notPruned = new ArrayList<MinimalDescription>();
			for(MinimalDescription description : collection.getObjects()) {
				double percentage = description.getPercentage(label);
				if(percentage<(expected-epsilon) || percentage>(expected+epsilon)) {
					notPruned.add(description);
				}
			}
			if(!notPruned.isEmpty()) {
				result.add(new MinimalDescriptionCollection(notPruned, collection.getDimension()));
			}
		}

		prunedMinimalDescription.deliver(result);
		passthroughPort.deliver(input);
	}
	
	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		
		types.add(new ParameterTypeString(PARAMETER_LABEL, "The label which distribution will be observed"));
		types.add(new ParameterTypeDouble(PARAMETER_PERCENTAGE, "The percentage which is expected for the given label", 0.0, 1.0, 0.33));
		types.add(new ParameterTypeDouble(PARAMETER_EPSILON, "The epsilon in which the percentage can differ from the expected value.", 0.0, 1.0, 0.2));
		
		return types;
	}

}
