package com.rapidminer.subspace;

import java.io.Serializable;

/**
 * An interval is defined for a single attribute and represents
 * an range where a point can be within or not.
 * 
 * @author Marcin Skirzynski
 */
public class Interval implements Comparable<Interval>, Serializable {
	private static final long serialVersionUID = 2678824587721771848L;

	/**
	 * The attribute where this interval is defined in 
	 */
	private final String attributeName;
	
	/**
	 * The left endpoint of the interval
	 */
	private final double left;
	
	/**
	 * the right endpoint of the interval
	 */
	private final double right;
	
	/**
	 * The density of this interval, i.e. the fraction of
	 * examples which are between the left and the right endpoints.
	 */
	private double density = Double.NaN;
	
	/**
	 * Creates an immutable interval object.
	 */
	public Interval(String attributeName, double left, double right) {
		if(attributeName==null || attributeName.isEmpty())
			throw new IllegalArgumentException("Attribute name can not be null or empty");
		if(left>right) 
			throw new IllegalArgumentException("Left endpoint can not be bigger than right endpoint in an interval");
		this.attributeName = attributeName;
		this.left = left;
		this.right = right;
	}

	/**
	 * Returns <code>true</code> if the given interval is 
	 * a neighbor if this one. This means that either the left is equal the
	 * right or the left is equal to the right. 
	 */
	public boolean neighborOf(Interval interval) {
		return 	neighbor(interval, Direction.LEFT) || neighbor(interval, Direction.RIGHT);
	}
	
	/**
	 * Returns true if the specified interval is the neighbor in the specified
	 * direction for this interval.
	 */
	public boolean neighbor(Interval interval, Direction direction) {
		switch (direction) {
		case LEFT:
			return getLeft()==interval.getRight();
		case RIGHT:
			return getRight()==interval.getLeft();
		default:
			return false;
		}
	}
	
	/**
	 * Returns <code>true</code> if the value is within this interval.
	 */
	public boolean contains(double value) {
		return value >= getLeft() && value < getRight();
	}

//	/**
//	 * Returns <code>true</code> if the specified interval is
//	 * the left neighbor of this interval.
//	 */
//	public boolean leftNeighbor(Interval interval) {
//		return getLeft()==interval.getRight();
//	}
//
//	/**
//	 * Returns <code>true</code> if the specified interval is
//	 * the right neighbor of this interval. 
//	 */
//	public boolean rightNeighbor(Interval interval) {
//		return getRight()==interval.getLeft();
//	}

	/**
	 * The attribute where this interval is defined
	 */
	public String getAttributeName() {
		return attributeName;
	}
	
	/**
	 * Returns the return endpoint in the specified direction
	 * for this interval.
	 */
	public double getEndpoint(Direction direction) {
		switch (direction) {
		case LEFT:
			return left;
		case RIGHT:
			return right;
		default:
			return Double.NaN;
		}
	}

	/**
	 * The left endpoint of this interval
	 * @return
	 */
	public double getLeft() {
		return getEndpoint(Direction.LEFT);
	}

	/**
	 * The right endpoint of this interval.
	 * @return
	 */
	public double getRight() {
		return getEndpoint(Direction.RIGHT);
	}

	/**
	 * Returns the fraction of the examples which are between the left
	 * and the right interval.
	 */
	public double getDensity() {
		return density;
	}

	/**
	 * Sets the fraction of the example which are between the left and the right interval. 
	 */
	public void setDensity(double density) {
		this.density = density;
	}
	
	/**
	 * Returns <code>true</code> if this interval is defined on for the same attribute.
	 */
	public boolean equalAttribute(Interval interval) {
		return interval.getAttributeName().equals(getAttributeName());
	}

	/**
	 * Simple string representation for this interval.
	 */
	@Override
	public String toString() {
		return "("+getLeft()+" < "+attributeName+" < "+getRight()+")";
	}
	
	/**
	 * Compares the two intervals but does not support overlapping intervals.
	 */
	@Override
	public int compareTo(Interval o) {
		return Double.compare(getLeft(), o.getLeft());
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + attributeName.hashCode();
		result = prime * result + (int) (left);
		result = prime * result + (int) (right);
		return result;
	}

	/**
	 * Two intervals are equal if their attribute name and their left and right
	 * endpoint are equal. The density will not be used.
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Interval other = (Interval) obj;

		if (!attributeName.equals(other.attributeName))
			return false;
		if(left!=other.left || right!=other.right)
			return false;
		return true;
	}
	
	public static enum Direction {
		LEFT, RIGHT;
	}

}
