package com.rapidminer.subspace;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.operator.ResultObjectAdapter;

/**
 * A minimal description of a single cluster.
 * 
 * @author Marcin Skirzynski
 */
public class MinimalDescription extends ResultObjectAdapter {
	private static final long serialVersionUID = 6846645491596163227L;
	
	/**
	 * Intern used string for all examples with no label.
	 */
	private static final String NO_LABEL = "";
	
	/**
	 * A minimal descriptions consists of a set of regions.
	 */
	private final Set<Region> regions;
	
	/**
	 * The attribute names for this cluster
	 */
	private final Set<String> attributes;

	/**
	 * A mapping from label to the number of contained
	 * examples with this label.
	 */
	private Map<String, Integer> containedExamples;
	
	/**
	 * The silhouette-coefficient, which is between
	 * -1 and +1. With a default value of -2 the coefficient
	 * will be ignored.
	 */
	private double silhouette = -2;
	
	/**
	 * The subspace where this cluster exists.
	 */
	private Subspace subspace;
	
	/**
	 * Creates a minimal description which consists of a set
	 * of regions.
	 */
	public MinimalDescription(Set<Region> regions) {
		if(regions==null || regions.isEmpty())
			throw new IllegalArgumentException("Can not build a minimal description of no regions");
		this.regions = new HashSet<Region>(regions);
		this.attributes = new HashSet<String>();
		if(regions!=null&&!regions.isEmpty()) {
			attributes.addAll(regions.iterator().next().getAttributes());
		}
	}
	
	/**
	 * Creates a minimal description which consists of a set
	 * of regions.
	 */
	public MinimalDescription(Region...regions) {
		if(regions==null || regions.length==0)
			throw new IllegalArgumentException("Can not build a minimal description of no regions");
		this.regions = new HashSet<Region>();
		for(Region region : regions)
			this.regions.add(region);
		this.attributes = new HashSet<String>();
		this.attributes.addAll(regions[0].getAttributes());
	}
		
	/**
	 * Creates a minimal description which consists of a set
	 * of regions.
	 */
	public MinimalDescription(Set<Region> regions, Subspace subspace) {
		this(regions);
		this.subspace = subspace;
	}

	/**
	 * A minimal descriptions consist of a set of regions which are returned
	 * with this method. 
	 */
	public Set<Region> getRegions() {
		return regions;
	}
	
	/**
	 * The subspace where this cluster is in.
	 */
	public Subspace getSubspace() {
		if(subspace==null) {
			subspace = new Subspace(attributes);
		}
		return subspace;
	}

	/**
	 * Sets the number of examples which are contained in this
	 * cluster with _NO_ label. 
	 */
	public void setContainedExamples(int examples) {
		setContainedExamples(examples, NO_LABEL);
	}
	
	/**
	 * Sets the number of examples which are contained in this
	 * cluster with the given label. 
	 */
	public void setContainedExamples(int examples, String label) {
		getContainedExamples().put(label, examples);
	}
	
	/**
	 * Adds one to the number of contained examples with _NO_ label.
	 */
	public void addContainedExample() {
		addContainedExample(NO_LABEL);
	}
	
	/**
	 * Adds one to the number of contained examples with the given label.
	 * @param label
	 */
	public void addContainedExample(String label) {
		if(!getContainedExamples().containsKey(label)) {
			getContainedExamples().put(label, 0);
		}
		getContainedExamples().put(label, getContainedExamples().get(label) + 1);
		
	}
	
	/**
	 * Null-safe Getter for the map of contained examples.
	 */
	private Map<String,Integer> getContainedExamples() {
		if(containedExamples==null)
			containedExamples = new HashMap<String, Integer>();
		return containedExamples;
	}

	/**
	 * Returns the absolute number of examples which are contained in
	 * this cluster regardless of the label it has.
	 */
	public int getAbsoluteCoverage() {
		int total = 0;
		for(int count : getContainedExamples().values()) {
			total += count;
		}
		return total;
	}
	
	/**
	 * Returns the absolute number of examples which are contained 
	 * in this cluster with the given label.
	 * @param label
	 * @return
	 */
	public int getAbsoluteCoverage(String label) {
		if(!getContainedExamples().containsKey(label))
			return 0;
		return getContainedExamples().get(label);
	}
	
	/**
	 * Returns the percentage of examples of the given label
	 * regarding all contained examples.  
	 */
	public double getPercentage(String label) {
		int absolute = getAbsoluteCoverage();
		int withLabel = getAbsoluteCoverage(label);
		
		return withLabel/(double)absolute;
	}
	
	/**
	 * Returns the silhouette-coefficient for this
	 * cluster.
	 */
	public double getSilhouette() {
		return silhouette;
	}

	/**
	 * Sets the silhouette-coeffizient for this cluster.
	 */
	public void setSilhouette(double silhouette) {
		this.silhouette = silhouette;
	}

	/**
	 * Returns <code>true</code> if the given example
	 * is contained in this cluster.
	 */
	public boolean contains(Example example) {
		for(Region region : this.getRegions()) {
			if(contains(example, region)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns <code>true</code> if the given example is within the
	 * given region. 
	 */
	private boolean contains(Example example, Region region) {
		for(Interval interval : region.getIntervals()) {
			String name = interval.getAttributeName();
			Attribute attribute = example.getAttributes().get(name);
			double value = example.getNumericalValue(attribute);
			if(!interval.contains(value)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * The number of dimensions of the subspace this cluster is in.
	 */
	public int getDimension() {
		return attributes.size();
	}
	
	/**
	 * Returns the short name of this cluster which will be listed in the RapidMiner
	 * GUI. 
	 */
	@Override
	public String getName() {
		StringBuilder builder = new StringBuilder();
		builder.append("MD[");
		for(String attribute : attributes) {
			if(builder.length()>30) {
				builder.append("...,");
				break;
			}
			builder.append(attribute);
			builder.append(',');
		}
		if(!attributes.isEmpty())
			builder.deleteCharAt(builder.length()-1);
		builder.append("]");
		return builder.toString();
	}
	
	/**
	 * Returns a long representation in DNF of this cluster which will be displayed as
	 * the result of this subspace cluster.
	 */
	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		for(Region region : regions) {
			builder.append(region.toString());
			builder.append("\n\tOR\n");
		}
		builder.setLength(builder.length()-5);
		if(!getContainedExamples().isEmpty()) {
			builder.append("\n\n");
			int total = 0;
			for(String label : getContainedExamples().keySet()) {
				int count = getContainedExamples().get(label);
				total += count;
				
				if(label.equals(NO_LABEL)) {
					builder.append("Cluster contains " + count + " examples with no label ("+getPercentage(label)*100+"%)\n");
				} else {
					builder.append("Cluster contains " + count + " example with the label '"+label+"' ("+getPercentage(label)*100+"%)\n");
				}
			}
			builder.append("-------\n");
			builder.append("Total number of containted examples: " + total);
			
		}
		
		if(silhouette>-2) {
			builder.append("\n\n");
			builder.append("Silhouette-coefficient is " + silhouette);
		}
		
		return builder.toString();
	}
	
}
