package com.rapidminer.subspace;

import java.util.List;

import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.IOObjectCollection;
import com.rapidminer.operator.preprocessing.statistics.SubspaceClusterStatistics;

/**
 * A collection of cluster descriptions for the same subspace.
 * 
 * @author Marcin Skirzynski
 */
public class MinimalDescriptionCollection extends IOObjectCollection<MinimalDescription> {
	private static final long serialVersionUID = -7800738507737312405L;

	/**
	 * The number of the dimesion
	 */
	private final int dimension;
	
	/**
	 * The jaccard index which is between 0 and 1. The default value
	 * of -1 will be ignored. Will be computed by {@link SubspaceClusterStatistics}.
	 */
	private double jaccard = -1;

	/**
	 * The rand index which is between 0 and 1. The default value
	 * of -1 will be ignored. Will be computed by {@link SubspaceClusterStatistics}.
	 */
	private double rand = -1;
	
	/**
	 * The coverage of this cluster, i.e. the percentage of examples
	 * this cluster covers. The default value
	 * of -1 will be ignored.
	 */
	private double coverage = -1;
	
	/**
	 * The silhouette-coefficient.
	 */
	private SilhouetteCoeffizient silhouette;
	
	/**
	 * The average silhouette-coefficient value.
	 */
	private double avgSilhouette = -2;
	
	/**
	 * The size of the example set to compute the coverage.
	 */
	private int exampleSetSize = -1;

	/**
	 * Creates a collection of minimal descriptions 
	 */
	public MinimalDescriptionCollection(List<MinimalDescription> descriptions, int dimension) {
		super(descriptions);
		this.dimension = dimension;
	}
	
	/**
	 * Calculates the silhouette of the contained examples without storing the examplesset.
	 */
	public void calculateSilhouette(ExampleSet exampleSet) {
		setSilhouette(new SilhouetteCoeffizient(getObjects()));
		getSilhouette().calculate(exampleSet);
		setAvgSilhouette(getSilhouette().getAvgSilhouette());
		
		for(MinimalDescription md : getObjects()) {
			md.setSilhouette(getSilhouette().getSilhouette(md));
		}
	}
	
	/**
	 * Returns a short name representation of this collection with
	 * all statistics which were computed.
	 */
	@Override
	public String getName() {
		StringBuilder builder = new StringBuilder();
		builder.append(size());
		builder.append(" cluster (");
		if(getDimension()>=0) 
			builder.append("dimension="+round(getDimension())+",");
		if(getAvgSilhouette()>-2)
			builder.append("silh="+round(getAvgSilhouette())+",");
		if(getCoverage()>=0)
			builder.append("cov="+round(getCoverage())+",");
		if(getJaccard()>=0)
			builder.append("jac="+round(getJaccard())+",");
		if(getRand()>=0)
			builder.append("rand="+round(getRand())+",");
		if(builder.charAt(builder.length()-1)==',')
			builder.deleteCharAt(builder.length()-1);
		builder.append(')');
		return builder.toString();
	}
	
	/**
	 * Rounds the given value to two digits after the comma. 
	 */
	private String round(double val) {
		return String.format("%.2f", val);
	}
	
	/**
	 * Returns the number of dimension of the subspace
	 * the clusters are in.
	 */
	public int getDimension() {
		return dimension;
	}

	/**
	 * Returns the jaccard index for the contained clusters. 
	 */
	public double getJaccard() {
		return jaccard;
	}

	/**
	 * Sets the jaccard index for the contained clusters. 
	 */
	public void setJaccard(double jaccard) {
		this.jaccard = jaccard;
	}

	/**
	 * Returns the jaccard index for the contained clusters. 
	 */
	public double getRand() {
		return rand;
	}

	/**
	 * Returns the rand index for the contained clusters. 
	 */
	public void setRand(double rand) {
		this.rand = rand;
	}

	/**
	 * Computes and returns the coverage of the clusters
	 */
	public double getCoverage() {
		if(coverage<0 && exampleSetSize>0) {
			int totalCover = 0;
			for(MinimalDescription md : getObjects()) {
				totalCover += md.getAbsoluteCoverage();
			}
			coverage = totalCover/(double)exampleSetSize;
		}
		return coverage;
	}

	/**
	 * Sets the coverage for this given clusters
	 * @param coverage
	 */
	public void setCoverage(double coverage) {
		this.coverage = coverage;
	}
	
	/**
	 * Sets the size of the example set to calculate the coverage.
	 * @param size
	 */
	public void setExampleSetSize(int size) {
		this.exampleSetSize = size;
	}
	
	/**
	 * Returns the size of the example set to calculate the coverage.
	 * @return
	 */
	public int getExampleSetSize() {
		return this.exampleSetSize;
	}

	/**
	 * Returns the average of all cluster silhouettes.
	 */
	public double getAvgSilhouette() {
		return avgSilhouette;
	}

	/**
	 * Sets the average of all cluster silhouettes.
	 */
	public void setAvgSilhouette(double avgSilhouette) {
		this.avgSilhouette = avgSilhouette;
	}

	public SilhouetteCoeffizient getSilhouette() {
		return silhouette;
	}

	public void setSilhouette(SilhouetteCoeffizient silhouette) {
		this.silhouette = silhouette;
	}


}
