package com.rapidminer.subspace;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.rapidminer.operator.ResultObjectAdapter;

/**
 * A regions is part of a cluster and is a rectangular unit in a subspace
 * of the whole data space.
 * 
 * @author Marcin Skirzynski
 */
public class Region extends ResultObjectAdapter {
	private static final long serialVersionUID = 345873933587487401L;
	
	/**
	 * The intervals which define the region
	 */
	private final Set<Interval> intervals;
	
	/**
	 * The boolean attributes for this region
	 */
	private final Set<String> booleans;
	
	/**
	 * The name of the attributes of this region.
	 */
	private final Set<String> attributes;
	
	/**
	 * Creates a region from the given units and their intervals which have to
	 * have a rectangle shape.
	 */
	public static Region createRegion(Subspace subspace, Set<Unit> units) {
		if(units==null || units.isEmpty())
			throw new IllegalArgumentException("Can not create region with no units");
		if(units.size()==1) {
			Unit unit = units.iterator().next();
			if(unit.isBooleanOnly()) {
				return new Region(unit.getBooleaAttributes());
			}
			return new Region(unit.getIntervals(), unit.getBooleaAttributes());
		}
		
		Set<Interval> regionIntervals = new HashSet<Interval>();
		Collection<String> booleanAttributes = new ArrayList<String>();
		
		for(Unit unit : units) {
			booleanAttributes.addAll(unit.getBooleaAttributes());
		}
		
		for(String attribute : subspace.getAttributeNames()) {
			List<Interval> intervals = new ArrayList<Interval>();
			for(Unit unit : units) {
				Interval interval = unit.getInterval(attribute);
				if(!intervals.contains(interval))
					intervals.add(unit.getInterval(attribute));
			}
			Collections.sort(intervals);
			
			// insanity check
			double lastRight = Double.NaN;
			for(Interval interval : intervals) {
				if(Double.isNaN(lastRight)) {
					lastRight = interval.getRight();
					continue;
				}
				if(lastRight!=interval.getLeft())  {
					throw new IllegalArgumentException("Units seem not to build a rectangle since for the attribute '"+
							attribute + "' there is a gap between " + lastRight + " and "+ interval.getLeft());
				} else {
					lastRight = interval.getRight();
				}
				
			}
			
			// Create new interval for the region
			double left = intervals.get(0).getLeft();
			double right = intervals.get(intervals.size()-1).getRight();
			regionIntervals.add(new Interval(attribute, left, right));
		}
		
		return new Region(regionIntervals, booleanAttributes);
	}
	
	/**
	 * Creates a region from the given intervals.
	 */
	public Region(Set<Interval> intervals) {
		if(intervals==null || intervals.isEmpty()) 
			throw new IllegalArgumentException("Can not create region with no intervals");
		this.intervals = intervals;
		this.booleans = Collections.emptySet();

		this.attributes = new HashSet<String>();
		for(Interval interval : intervals) {
			this.attributes.add(interval.getAttributeName());
		}
	}
	
	/**
	 * Creates a region from the given intervals and with the given boolean
	 * attributes.
	 */
	public Region(Set<Interval> intervals, Collection<String> booleans) {
		if(intervals==null || intervals.isEmpty()) 
			throw new IllegalArgumentException("Can not create region with no intervals");
		this.intervals = intervals;
		this.booleans = new HashSet<String>(booleans);
		this.attributes = new HashSet<String>();
		for(Interval interval : intervals) {
			this.attributes.add(interval.getAttributeName());
		}
	}
	
	/**
	 * Returns a boolean only region. 
	 */
	public Region(Collection<String> booleans) {
		this.intervals = Collections.emptySet();
		this.booleans = new HashSet<String>(booleans);
		this.attributes = Collections.emptySet();
	}

	/**
	 * Returns the intervals which define this region.
	 */
	public Set<Interval> getIntervals() {
		return intervals;
	}

	/**
	 * Returns the name of the attributes for which this
	 * region is defined.
	 */
	public Set<String> getAttributes() {
		return attributes;
	}

	/**
	 * Returns a long representation of this region in DNF.
	 */
	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		for(Interval interval : intervals) {
			builder.append('(');
			builder.append(interval.getLeft());
			builder.append('<');
			builder.append(interval.getAttributeName());
			builder.append('<');
			builder.append(interval.getRight());
			builder.append(')');
			builder.append("AND");
		}
		for(String attribute : booleans) {
			builder.append('(');
			builder.append(attribute);
			builder.append(')');
			builder.append("AND");
		}
		return builder.substring(0,builder.length()-3);
	}

}
