package com.rapidminer.subspace;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.tools.math.similarity.DistanceMeasure;
import com.rapidminer.tools.math.similarity.numerical.EuclideanDistance;

/**
 * Computes the silhouette-coefficient for a set of minimal
 * descriptions.
 * 
 * @author Marcin Skirzynski
 *
 */
public class SilhouetteCoeffizient implements Serializable{
	private static final long serialVersionUID = 1574317309576222124L;

	/**
	 * The minimal descriptions for which the coefficient will
	 * be calculated.
	 */
	private final List<MinimalDescription> mds;
	
	/**
	 * The distance measure which will be used for the calculations.
	 */
	private final DistanceMeasure distance;
	
	/**
	 * All silhouettes of all examples for all clusters.
	 */
	private final Map<MinimalDescription, List<Double>> coeffizient;
	
	/**
	 * Creates object for calculation of the coefficient for the given descriptions and
	 * the euclidean distance measure. 
	 */
	public SilhouetteCoeffizient(List<MinimalDescription> mds) {
		this(mds, new EuclideanDistance());
	}
	
	/**
	 * Creates object for calculation of the coefficient for the given descriptions and
	 * the given distance measure. 
	 */
	public SilhouetteCoeffizient(List<MinimalDescription> mds, DistanceMeasure distance) {
		this.mds = mds;
		this.distance = distance;
		this.coeffizient = new HashMap<MinimalDescription, List<Double>>();
	}
	
	/**
	 * Calculates the coefficient for the given examples
	 */
	public void calculate(ExampleSet exampleSet) {
		
		Map<MinimalDescription, Set<Example>> assignments = getAssignments(exampleSet);
		for(MinimalDescription md1 : assignments.keySet()) {
			if(md1==null)
				continue;
			
			coeffizient.put(md1, new ArrayList<Double>());
			for(Example example : assignments.get(md1)) {

				
				double min = averageDistance(example, assignments.get(null));
				for(MinimalDescription md2 : getDifference(md1, mds)) {
					double outerAvgDistance = averageDistance(example, assignments.get(md2));
					if(outerAvgDistance<min)
						min = outerAvgDistance;
				}
				
				double a = averageDistance(example, assignments.get(md1));
				double b = min;
				
				double exampleSilhouette = (b-a)/(Math.max(a, b));
				coeffizient.get(md1).add(exampleSilhouette);
				
			}
		}
		
	}
	
	/**
	 * Returns the coefficient for the given description.
	 */
	public double getSilhouette(MinimalDescription md) {
		double sum = 0.0;
		for(double silhouette : coeffizient.get(md)) {
			sum += silhouette;
		}
		return sum/(double)coeffizient.get(md).size();
	}
	
	/**
	 * Returns the average silhouette coefficient for all clusters.
	 */
	public double getAvgSilhouette() {
		double sum = 0.0;
		for(MinimalDescription md : coeffizient.keySet()) {
			if(md==null)
				continue;
			sum += getSilhouette(md);
		}
		return sum/(double)coeffizient.keySet().size();
	}
	
	/**
	 * Returns a map where all examples are assigned to the minimal description
	 * which contain them.
	 */
	private Map<MinimalDescription, Set<Example>> getAssignments(ExampleSet exampleSet) { 
		Map<MinimalDescription, Set<Example>> result = new HashMap<MinimalDescription, Set<Example>>();
		
		for(Example example : exampleSet) {
			MinimalDescription md = getMinimalDescription(example);
			if(md!=null) {
				if(!result.containsKey(md)) {
					result.put(md, new HashSet<Example>());
				}
				result.get(md).add(example);
			} else {
				if(!result.containsKey(null)) {
					result.put(null, new HashSet<Example>());
				}
				result.get(null).add(example);
			}
		}
		
		return result;
	}
	
	/**
	 * Returns the average distance between the given example and the other examples. 
	 */
	private double averageDistance( Example example, Set<Example> examples) {
		double sum = 0.0;
		for(Example other : examples) {
			sum += distance(example, other);
		}
		return sum/(double)examples.size();
	}
	
	/**
	 * Calculates the distance for the attributes used
	 * in the subspace. 
	 */
	private double distance(Example e1, Example e2) {
		Subspace subspace = mds.get(0).getSubspace();
		double[] values1 = new double[subspace.size()];
		double[] values2 = new double[subspace.size()];
	
		int i=0;
		for(String attributeName : subspace.getAttributeNames()) {
			Attribute attribute = e1.getAttributes().get(attributeName);
			values1[i] = e1.getNumericalValue(attribute);
			values2[i] = e2.getNumericalValue(attribute);
			i++;
		}
		return distance.calculateDistance(values1, values2);
	}

	/**
	 * Returns the description where this example is contained.
	 */
	private MinimalDescription getMinimalDescription(Example example) {
		for(MinimalDescription md : mds) {
			if(md.contains(example)) {
				return md;
			}
		}
		return null;
	}
	
	/**
	 * Returns a new list of descriptions without the given description in it.
	 */
	private List<MinimalDescription> getDifference(MinimalDescription md, List<MinimalDescription> mds) {
		List<MinimalDescription> result = new ArrayList<MinimalDescription>(mds);
		result.remove(md);
		return result;
	}

}
