package com.rapidminer.subspace;

import java.util.Collections;
import java.util.Set;

import com.rapidminer.operator.ResultObjectAdapter;
import com.rapidminer.subspace.Interval.Direction;


/**
 * A subspace cluster is a set of dense units in a specific subspace.
 * 
 * @author Marcin Skirzynski
 *
 */
public class SubspaceCluster extends ResultObjectAdapter {
	private static final long serialVersionUID = 6847648010708042248L;

	/**
	 * The subspace of this cluster.
	 */
	private final Subspace subspace;

	/**
	 * The set of connected units which form this cluster in the given subspace.
	 */
	private final Set<Unit> units;
	
	/**
	 * Constructs an immutable subspace clustering object.
	 */
	public SubspaceCluster(Subspace subspace, Set<Unit> units) {
		this.subspace = subspace;
		this.units = units;
	}

	/**
	 * Returns an unmodifiable set of the connected dense units which
	 * form this cluster.
	 */
	public Set<Unit> getUnits() {
		return Collections.unmodifiableSet(units);
	}
	
	/**
	 * Returns <code>true</code> if this cluster consists just of one unit.
	 */
	public boolean isSingleUnitCluster() {
		return units.size()==1;
	}
	
	/**
	 * Returns the right unit of this specified unit for the specified attribute
	 * which is also contained in this cluster and <code>null</code> otherwise.
	 * 
	 *  If this unit is not within this cluster <code>null</code> will be returned as well.
	 */
	public Unit getNeighbor(Unit unit, Direction direction, String attribute) {
		if(!units.contains(unit))
			return null;
		for(Unit potentialNeighbor : units) {
			boolean isNeighbor = unit.neighbor(potentialNeighbor, direction, attribute); 
			if(isNeighbor) 
				return potentialNeighbor;
		}
		return null;
	}

	/**
	 * The subspace where this is a cluster.
	 */
	public Subspace getSubspace() {
		return subspace;
	}
	
	/**
	 * Simple string representation of this cluster. 
	 */
	@Override
	public String toString() {
		return "SubspaceCluster ("+units.size()+" units)"+"[" + units + "]";
	}

}
