package com.rapidminer.subspace;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.rapidminer.subspace.Interval.Direction;


/**
 * A unit is a rectangular object within a subspace. It is
 * defined by a set of intervals for every attributes
 * in this subspace.
 * 
 * @author Marcin Skirzynski
 *
 */
public class Unit implements Serializable {
	private static final long serialVersionUID = 3075661622965972867L;

	/**
	 * Intervals which define this unit. Every interval represents
	 * a range for a single attribute.
	 */
	private final Set<Interval> intervals;
	
	/**
	 * Maps the name of an attribute to the interval for this attribute.
	 */
	private final Map<String, Interval> attributes;

	/**
	 * The density of this unit which is the fraction of examples
	 * within this unit when it was built.
	 */
	private final double density;
	
	/**
	 * List of boolean attribute names which are also included 
	 */
	private Collection<String> booleans = Collections.emptyList();
	

	public Unit(Collection<Interval> intervals, double density) {
		this.intervals = new HashSet<Interval>(intervals);
		this.density = density;
		this.attributes = new HashMap<String, Interval>();
		for(Interval interval : intervals) {
			String attribute = interval.getAttributeName();
			if(this.attributes.keySet().contains(attribute))
				throw new IllegalArgumentException("Unit created with two intervals for the same attrubte '" + attribute + "'");
			this.attributes.put(attribute, interval);
		}
	}
	
	/**
	 * Two k-dimensional units have a common face if there are
	 * k-1 attributes for which the intervals are equal and one attribute
	 * where the intervals are neighbors. 
	 */
	public boolean hasCommonFace(Unit unit) {
		if(!equalAttributes(unit))
			return false;
		Collection<String> attributes = getAttributes();
		boolean neighborFound = false;
		for(String attribute : attributes) {
			Interval i1 = getInterval(attribute);
			Interval i2 = unit.getInterval(attribute);
			if(i1.neighborOf(i2)) {
				// There can only be one attribute where the intervals are neighbor
				if(neighborFound) 
					return false;
				neighborFound = true;
				continue;
			}
			if(!i1.equals(i2))
				return false;
		}
		return true;
	}
	
	/**
	 * Returns the interval for the given attribute.
	 */
	public Interval getInterval(String attribute) {
		return attributes.get(attribute);
	}

	/**
	 * Returns an unmodifiable set of the intervals of this unit.
	 */
	public Set<Interval> getIntervals() {
		return Collections.unmodifiableSet(intervals);
	}

	/**
	 * Returns an unmodifiable set of the attributes of this set.
	 */
	public Set<String> getAttributes() {
		return Collections.unmodifiableSet(attributes.keySet());
	}

	/**
	 * Creates a new immutable subspace in which this
	 * unit is defined.
	 */
	public Subspace getSubspace() {
		return new Subspace(getAttributes());
	}

	/**
	 * Returns the density of this unit which is the fraction
	 * of examples which are contained in this unit (i.e. within
	 * the intervals).
	 */
	public double getDensity() {
		return density;
	}
	

	/**
	 * Sets the boolean attributes which are also include in this unit.
	 */
	public void setBooleanAttributes(Collection<String> booleanAttributes) {
		booleans = booleanAttributes;
	}
	
	/**
	 * Sets the boolean attributes which are also include in this unit.
	 */
	public Collection<String> getBooleaAttributes() {
		return Collections.unmodifiableCollection(booleans);
	}
	
	/**
	 * Returns <code>true</code> if this unit consits just of 
	 * boolean attributes.
	 */
	public boolean isBooleanOnly() {
		return intervals.size()==0 && booleans.size()>0;
	}
	
	/**
	 * Returns <code>true</code> if the specified unit is the neighbor in the given direction 
	 * on the specified attribute.
	 */
	public boolean neighbor(Unit unit, Direction direction, String attribute) {
		if(intervals.size()!=unit.intervals.size())
			return false;
		boolean neighborOnAttribute = false;
		for(Interval thisInterval : intervals) {

			String currentAttribute = thisInterval.getAttributeName();
			Interval thatInterval = unit.getInterval(currentAttribute);
			
			if(currentAttribute.equals(attribute)) {
				if(thisInterval.neighbor(thatInterval, direction)) {
					neighborOnAttribute = true;
				} else {
					return false;
				}
			} else {
				if(!thisInterval.equals(thatInterval)) {
					return false;
				}
			}
		}
		return neighborOnAttribute;
	}

	/**
	 * Simple string representation for this unit
	 */
	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		for(Interval interval : intervals) {
			builder.append('(');
			builder.append(interval.getLeft());
			builder.append('<');
			builder.append(interval.getAttributeName());
			builder.append('<');
			builder.append(interval.getRight());
			builder.append(')');
			builder.append("AND");
		}
		for(String attribute : booleans) {
			builder.append('(');
			builder.append(attribute);
			builder.append(')');
			builder.append("AND");
		}
		return "[" + builder.substring(0,builder.length()-3) + "]";
	}
	
	/**
	 * Returns <code>true</code> if the given unit has the
	 * same attributes as this unit.
	 */
	public boolean equalAttributes(Unit unit) {
		Set<String> thatKeyset = unit.attributes.keySet();
		Set<String> thisKeyset = this.attributes.keySet();
		
		if(thatKeyset.size()!=thisKeyset.size())
			return false;
		
		return thatKeyset.containsAll(thisKeyset);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((attributes == null) ? 0 : attributes.hashCode());
		long temp;
		temp = Double.doubleToLongBits(density);
		result = prime * result + (int) (temp ^ (temp >>> 32));
		result = prime * result
				+ ((intervals == null) ? 0 : intervals.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Unit other = (Unit) obj;
		if (attributes == null) {
			if (other.attributes != null)
				return false;
		} else if (!attributes.equals(other.attributes))
			return false;
		if (Double.doubleToLongBits(density) != Double
				.doubleToLongBits(other.density))
			return false;
		if (intervals == null) {
			if (other.intervals != null)
				return false;
		} else if (!intervals.equals(other.intervals))
			return false;
		return true;
	}


}
